from abc import ABC, abstractmethod

import numpy as np
import torch
from typing_extensions import Protocol

from l5kit.data import get_agents_slice_from_frames
from l5kit.evaluation import error_functions
from l5kit.evaluation import metrics as l5metrics
from l5kit.rasterization import EGO_EXTENT_LENGTH, EGO_EXTENT_WIDTH
from l5kit.simulation.unroll import SimulationOutputCLE, TrajectoryStateIndices


class SupportsMetricCompute(Protocol):
    """Protocol supporting the computation method for metrics."""
    metric_name: str

    @abstractmethod
    def compute(self, simulation_output: SimulationOutputCLE) -> torch.Tensor:
        """The compute method sould return the result of the metric
        computed at every frame of the scene.

        :param simulation_output: the output from the closed-loop simulation
        :returns: a tensor with the result of the metric per frame
        """
        raise NotImplementedError


class CollisionMetricBase(ABC):
    """This is the abstract base class for the collision metric.

    :param collision_type: the type of collision to compute
    """
    @abstractmethod
    def __init__(self, collision_type: l5metrics.CollisionType) -> None:
        self.collision_type = collision_type

    def _compute_frame(self, simulated_agent_frame: np.ndarray,
                       simulated_frame_ego_state: torch.Tensor) -> float:
        """Detects collision per frame of the scene.

        :param observed_frame: the ground-truth frame
        :param simulated_frame_ego_state: ego state from the simulated frame,
                                          this is a 1D array with the frame
                                          ego state
        :returns: metric result for the frame, where 1 means a collision,
                  and 0 otherwise.
        """
        simulated_centroid = simulated_frame_ego_state[:TrajectoryStateIndices.Y + 1].cpu().numpy()
        simulated_angle = simulated_frame_ego_state[TrajectoryStateIndices.THETA].cpu().numpy()
        simulated_extent = np.r_[EGO_EXTENT_LENGTH, EGO_EXTENT_WIDTH]
        collision_ret = l5metrics.detect_collision(simulated_centroid, simulated_angle,
                                                   simulated_extent, simulated_agent_frame)
        if collision_ret is not None:
            if collision_ret[0] == self.collision_type:
                return 1.0

        return 0.0

    def compute(self, simulation_output: SimulationOutputCLE) -> torch.Tensor:
        """Compute the metric on all frames of the scene.

        :param simulation_output: the output from the closed-loop simulation
        :returns: collision per frame (a 1D array with the same size of the
                  number of frames, where 1 means a colision, 0 otherwise)
        """
        simulated_scene_ego_state = simulation_output.simulated_ego_states
        simulated_agents = simulation_output.simulated_agents
        simulated_egos = simulation_output.simulated_ego

        if len(simulated_agents) < len(simulated_scene_ego_state):
            raise ValueError("More simulated timesteps than observed.")

        num_frames = simulated_scene_ego_state.size(0)
        metric_results = torch.zeros(num_frames, device=simulated_scene_ego_state.device)
        for frame_idx in range(num_frames):
            simulated_ego_state_frame = simulated_scene_ego_state[frame_idx]
            simulated_ego_frame = simulated_egos[frame_idx]
            simulated_agent_frame = simulated_agents[get_agents_slice_from_frames(simulated_ego_frame)]
            result = self._compute_frame(simulated_agent_frame, simulated_ego_state_frame)
            metric_results[frame_idx] = result
        return metric_results


class CollisionFrontMetric(CollisionMetricBase):
    """Computes the front collision metric."""
    metric_name = "collision_front"

    def __init__(self) -> None:
        super().__init__(l5metrics.CollisionType.FRONT)


class CollisionRearMetric(CollisionMetricBase):
    """Computes the rear collision metric."""
    metric_name = "collision_rear"

    def __init__(self) -> None:
        super().__init__(l5metrics.CollisionType.REAR)


class CollisionSideMetric(CollisionMetricBase):
    """Computes the side collision metric."""
    metric_name = "collision_side"

    def __init__(self) -> None:
        super().__init__(l5metrics.CollisionType.SIDE)


class DisplacementErrorMetric(SupportsMetricCompute):
    """Displacement error computes the elementwise distance from the
    simulated trajectory and the observed trajectory.

    :param error_function: error function to compute distance
    """
    metric_name = "displacement_error"

    def __init__(self, error_function: error_functions.ErrorFunction) -> None:
        self.error_function = error_function

    def compute(self, simulation_output: SimulationOutputCLE) -> torch.Tensor:
        """Compute the metric on all frames of the scene.

        :param simulation_output: the output from the closed-loop simulation
        :returns: distance per frame [Shape: N, where N = timesteps]
        """
        simulated_scene_ego_state = simulation_output.simulated_ego_states
        simulated_centroid = simulated_scene_ego_state[:, :2]  # [Timesteps, 2]
        observed_ego_states = simulation_output.recorded_ego_states[:, :2]  # [Timesteps, 2]

        if len(observed_ego_states) < len(simulated_centroid):
            raise ValueError("More simulated timesteps than observed.")

        # Don't have simulation for all steps, have to clip it
        observed_ego_states_fraction = observed_ego_states[:len(simulated_centroid)]

        error = self.error_function(simulated_centroid, observed_ego_states_fraction)
        return error


class DisplacementErrorL2Metric(DisplacementErrorMetric):
    """Displacement error calculated with euclidean distance."""
    metric_name = "displacement_error_l2"

    def __init__(self) -> None:
        super().__init__(error_functions.l2_error)


class DistanceToRefTrajectoryMetric(SupportsMetricCompute):
    """Distance to reference trajectory metric. This metric will compute
    the distance from the predicted centroid to the closest waypoint
    in the reference trajectory.

    .. note::  Please note that this metric is different than the displacement
               error because it is taking into consideration the entire
               reference trajectory at each point of the simulated trajectory.

    :param scene_fraction: fraction of the simulated scene used to
                           evaluate against the reference trajectory. This
                           fraction should be between 0.0 and 1.0.
    """
    metric_name = "distance_to_reference_trajectory"

    def __init__(self, scene_fraction: float = 0.8) -> None:
        # Check fraction range
        if scene_fraction < 0.0 or scene_fraction > 1.0:
            raise ValueError("'screne_fraction' should be between 0.0 and 1.0.")

        self.scene_fraction = scene_fraction

    def compute(self, simulation_output: SimulationOutputCLE) -> torch.Tensor:
        """Compute the metric on all frames of the scene.

        :param simulation_output: the output from the closed-loop simulation
        :returns: distance to reference trajectory per
                  frame [Shape: N, where N = timesteps]
        """
        # Shape = [Timesteps, 7]
        simulated_scene_ego_state = simulation_output.simulated_ego_states
        simulated_centroid = simulated_scene_ego_state[:, :2]  # [Timesteps, 2]
        observed_ego_states = simulation_output.recorded_ego_states[:, :2]

        if len(observed_ego_states) < len(simulated_centroid):
            raise ValueError("More simulated timesteps than observed.")

        # Trim the simulated trajectory to have a specified fraction
        simulated_fraction_length = int(len(simulated_centroid) * self.scene_fraction)
        simulated_centroid_fraction = simulated_centroid[0:simulated_fraction_length]

        observed_ego_states = observed_ego_states.unsqueeze(0)
        distance = l5metrics.distance_to_reference_trajectory(simulated_centroid_fraction,
                                                              observed_ego_states)
        return distance


class SimulatedDrivenMilesMetric:
    """This metric will compute the driven miles per frame for the simulated
    trajectory (as opposed to the one in the log replay)."""
    metric_name = "simulated_driven_miles"
    METER_TO_MILES = 0.000621371

    def compute(self, simulation_output: SimulationOutputCLE) -> torch.Tensor:
        """Compute the metric on all frames of the scene.

        :param simulation_output: the output from the closed-loop simulation
        :returns: driven miles per each frame
        """
        simulated_scene_ego_state = simulation_output.simulated_ego_states
        simulated_centroid = simulated_scene_ego_state[:, :2]  # [Timesteps, 2]
        simulated_centroid = simulated_centroid.to(torch.float64)

        drive_meters = torch.linalg.norm(simulated_centroid[1:] - simulated_centroid[0:-1], dim=1)
        pad = torch.as_tensor([0.], device=simulated_centroid.device)
        pad_drive_meters = torch.cat((pad, drive_meters))
        driven_miles = pad_drive_meters * self.METER_TO_MILES
        return driven_miles


class ReplayDrivenMilesMetric:
    """This metric will compute the driven miles per frame for the observed
    trajectory, the one in the log replay (as opposed to the one simulated)."""
    metric_name = "replay_driven_miles"
    METER_TO_MILES = 0.000621371

    def compute(self, simulation_output: SimulationOutputCLE) -> torch.Tensor:
        """Compute the metric on all frames of the scene.

        :param simulation_output: the output from the closed-loop simulation
        :returns: driven miles per each frame
        """
        observed_ego_states_centroid = simulation_output.recorded_ego_states[:, :2]  # [Timesteps, 2]
        observed_ego_states_centroid = observed_ego_states_centroid.to(torch.float64)

        drive_meters = \
            torch.linalg.norm(observed_ego_states_centroid[1:]
                              - observed_ego_states_centroid[0:-1], dim=1)
        pad = torch.as_tensor([0.], device=observed_ego_states_centroid.device)
        pad_drive_meters = torch.cat((pad, drive_meters))
        driven_miles = pad_drive_meters * self.METER_TO_MILES
        return driven_miles


class YawErrorMetric(SupportsMetricCompute):
    """Yaw error computes the difference between the
    simulated trajectory yaw and the observed trajectory yaw.

    :param error_function: error function to compute distance
    """
    metric_name = "yaw_error"

    def __init__(self, error_function: error_functions.ErrorFunction = error_functions.closest_angle_error) -> None:
        self.error_function = error_function

    def compute(self, simulation_output: SimulationOutputCLE) -> torch.Tensor:
        """Compute the metric on all frames of the scene.

        :param simulation_output: the output from the closed-loop simulation
        :returns: distance per frame [Shape: N, where N = timesteps]
        """
        simulated_scene_ego_state = simulation_output.simulated_ego_states
        simulated_yaws = simulated_scene_ego_state[:, 2:3]  # [Timesteps,]
        observed_ego_yaws = simulation_output.recorded_ego_states[:, 2:3]  # [Timesteps,]

        if len(observed_ego_yaws) < len(simulated_yaws):
            raise ValueError("More simulated timesteps than observed.")

        # Don't have simulation for all steps, have to clip it
        observed_ego_yaws_fraction = observed_ego_yaws[:len(simulated_yaws)]

        error = self.error_function(simulated_yaws, observed_ego_yaws_fraction)
        return error


class YawErrorCAMetric(YawErrorMetric):
    """Yaw error calculated with closest angle."""
    metric_name = "yaw_error_closest_angle"

    def __init__(self) -> None:
        super().__init__(error_functions.closest_angle_error)



import numpy as np

from ..geometry import vertical_flip


class RenderContext:
    def __init__(
            self,
            raster_size_px: np.ndarray,
            pixel_size_m: np.ndarray,
            center_in_raster_ratio: np.ndarray,
            set_origin_to_bottom: bool,
    ) -> None:
        """
        This class stores render context information (raster size, pixel size, raster center / principle point) and
        it computes a transformation matrix (raster_from_local) to transform a local coordinates into raster ones.
        (0,0) in local will end up in the center of the raster (specified by combining `raster_size_px` and
        `center_in_raster_ratio`).

        Args:
            raster_size_px (Tuple[int, int]): Raster size in pixels
            pixel_size_m (np.ndarray): Size of one pixel in the real world, meter per pixel
            center_in_raster_ratio (np.ndarray): Where to center the local pose in the raster. [0.5,0.5] would be in
                the raster center, [0, 0] is bottom left.
        """

        if pixel_size_m[0] != pixel_size_m[1]:
            raise NotImplementedError("No support for non squared pixels yet")

        self.raster_size_px = raster_size_px
        self.pixel_size_m = pixel_size_m
        self.center_in_raster_ratio = center_in_raster_ratio
        self.set_origin_to_bottom = set_origin_to_bottom

        scaling = 1.0 / pixel_size_m  # scaling factor from world to raster space [pixels per meter]
        center_in_raster_px = center_in_raster_ratio * raster_size_px
        self.raster_from_local = np.array(
            [[scaling[0], 0, center_in_raster_px[0]], [0, scaling[1], center_in_raster_px[1]], [0, 0, 1]]
        )
        if set_origin_to_bottom:
            self.raster_from_local = vertical_flip(self.raster_from_local, self.raster_size_px[1])

    def raster_from_world(self, position_m: np.ndarray, angle_rad: float) -> np.ndarray:
        """
        Return a matrix to convert a pose in world coordinates into raster coordinates

        Args:
            render_context (RenderContext): the context for rasterisation
            position_m (np.ndarray): XY position in world coordinates
            angle_rad (float): rotation angle in world coordinates

        Returns:
            (np.ndarray): a transformation matrix from world coordinates to raster coordinates
        """
        # Compute pose from its position and rotation
        pose_in_world = np.array(
            [
                [np.cos(angle_rad), -np.sin(angle_rad), position_m[0]],
                [np.sin(angle_rad), np.cos(angle_rad), position_m[1]],
                [0, 0, 1],
            ]
        )

        pose_from_world = np.linalg.inv(pose_in_world)
        raster_from_world = self.raster_from_local @ pose_from_world
        return raster_from_world



from .gif import write_gif
from .utils import (draw_reference_trajectory, draw_trajectory, PREDICTED_POINTS_COLOR, REFERENCE_TRAJ_COLOR,
                    TARGET_POINTS_COLOR)
from .video import write_video


__all__ = [
    "write_gif",
    "write_video",
    "draw_reference_trajectory",
    "draw_trajectory",
    "TARGET_POINTS_COLOR",
    "PREDICTED_POINTS_COLOR",
    "REFERENCE_TRAJ_COLOR",
]



from abc import ABC, abstractmethod
from typing import Tuple

import numpy as np


class Perturbation(ABC):
    @abstractmethod
    def perturb(self, history_frames: np.ndarray, future_frames: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Args:
            history_frames (np.ndarray): array of past frames
            future_frames (np.ndarray): array of future frames
            kwargs: optional extra arguments for the specific perturber

        Returns:
            history_frames (np.ndarray): array of past frames with perturbation applied
            future_frames (np.ndarray): array of future frames with perturbation applied
        """
        pass



import warnings

import cv2
import numpy as np

from l5kit.configs.config import load_metadata

from ..data import DataManager
from .box_rasterizer import BoxRasterizer
from .rasterizer import Rasterizer
from .render_context import RenderContext
from .sat_box_rasterizer import SatBoxRasterizer
from .satellite_rasterizer import SatelliteRasterizer
from .sem_box_rasterizer import SemBoxRasterizer
from .semantic_rasterizer import SemanticRasterizer
from .stub_rasterizer import StubRasterizer


def _load_satellite_map(image_key: str, data_manager: DataManager) -> np.ndarray:
    """Loads image from given key.

    Args:
        image_key (str): key to the image (e.g. ``maps/my_satellite_image.png``)
        data_manager (DataManager): DataManager used for requiring files

    Returns:
        np.ndarry: Image
    """

    image_path = data_manager.require(image_key)
    image = cv2.imread(image_path)[..., ::-1]  # BGR->RGB
    if image is None:
        raise Exception(f"Failed to load image from {image_path}")

    return image


def get_hardcoded_world_to_ecef() -> np.ndarray:  # TODO remove when new dataset version is available
    """
    Return and hardcoded world_to_ecef matrix for dataset V1.0

    Returns:
        np.ndarray: 4x4 matrix
    """
    warnings.warn(
        "!!dataset metafile not found!! the hard-coded matrix will be loaded.\n"
        "This will be deprecated in future releases",
        PendingDeprecationWarning,
        stacklevel=3,
    )

    world_to_ecef = np.asarray(
        [
            [8.46617444e-01, 3.23463078e-01, -4.22623402e-01, -2.69876744e06],
            [-5.32201938e-01, 5.14559352e-01, -6.72301845e-01, -4.29315158e06],
            [-3.05311332e-16, 7.94103464e-01, 6.07782600e-01, 3.85516476e06],
            [0.00000000e00, 0.00000000e00, 0.00000000e00, 1.00000000e00],
        ],
        dtype=np.float64,
    )
    return world_to_ecef


def get_hardcoded_ecef_to_aerial() -> np.ndarray:  # TODO remove when new dataset version is available
    """
    Return and hardcoded ecef_to_aerial matrix for dataset V1.0

    Returns:
        np.ndarray: 4x4 matrix
    """
    warnings.warn(
        "!!dataset metafile not found!! the hard-coded matrix will be loaded.\n"
        "This will be deprecated in future releases",
        PendingDeprecationWarning,
        stacklevel=3,
    )

    ecef_to_aerial = np.asarray(
        [
            [-0.717416495, -1.14606296, -1.62854453, -572869.824],
            [1.80065798, -1.08914046, -0.0287877303, 300171.963],
            [0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 1.0],
        ],
        dtype=np.float64,
    )
    return ecef_to_aerial


def build_rasterizer(cfg: dict, data_manager: DataManager) -> Rasterizer:
    """Factory function for rasterizers, reads the config, loads required data and initializes the correct rasterizer.

    Args:
        cfg (dict): Config.
        data_manager (DataManager): Datamanager that is used to require files to be present.

    Raises:
        NotImplementedError: Thrown when the ``map_type`` read from the config doesn't have an associated rasterizer
        type in this factory function. If you have custom rasterizers, you can wrap this function in your own factory
        function and catch this error.

    Returns:
        Rasterizer: Rasterizer initialized given the supplied config.
    """
    raster_cfg = cfg["raster_params"]
    map_type = raster_cfg["map_type"]
    dataset_meta_key = raster_cfg["dataset_meta_key"]

    render_context = RenderContext(
        raster_size_px=np.array(raster_cfg["raster_size"]),
        pixel_size_m=np.array(raster_cfg["pixel_size"]),
        center_in_raster_ratio=np.array(raster_cfg["ego_center"]),
        set_origin_to_bottom=raster_cfg["set_origin_to_bottom"],
    )

    filter_agents_threshold = raster_cfg["filter_agents_threshold"]
    history_num_frames = cfg["model_params"]["history_num_frames"]
    render_ego_history = cfg["model_params"]["render_ego_history"]

    if map_type in ["py_satellite", "satellite_debug"]:
        sat_image = _load_satellite_map(raster_cfg["satellite_map_key"], data_manager)

        try:
            dataset_meta = load_metadata(data_manager.require(dataset_meta_key))
            world_to_ecef = np.array(dataset_meta["world_to_ecef"], dtype=np.float64)
            ecef_to_aerial = np.array(dataset_meta["ecef_to_aerial"], dtype=np.float64)

        except (KeyError, FileNotFoundError):  # TODO remove when new dataset version is available
            world_to_ecef = get_hardcoded_world_to_ecef()
            ecef_to_aerial = get_hardcoded_ecef_to_aerial()

        world_to_aerial = np.matmul(ecef_to_aerial, world_to_ecef)
        if map_type == "py_satellite":
            return SatBoxRasterizer(
                render_context, filter_agents_threshold, history_num_frames, sat_image, world_to_aerial,
                render_ego_history=render_ego_history
            )
        else:
            return SatelliteRasterizer(render_context, sat_image, world_to_aerial)

    elif map_type in ["py_semantic", "semantic_debug"]:
        semantic_map_filepath = data_manager.require(raster_cfg["semantic_map_key"])
        try:
            dataset_meta = load_metadata(data_manager.require(dataset_meta_key))
            world_to_ecef = np.array(dataset_meta["world_to_ecef"], dtype=np.float64)
        except (KeyError, FileNotFoundError):  # TODO remove when new dataset version is available
            world_to_ecef = get_hardcoded_world_to_ecef()
        if map_type == "py_semantic":
            return SemBoxRasterizer(
                render_context, filter_agents_threshold, history_num_frames, semantic_map_filepath, world_to_ecef,
                render_ego_history=render_ego_history
            )
        else:
            return SemanticRasterizer(render_context, semantic_map_filepath, world_to_ecef)

    elif map_type == "box_debug":
        return BoxRasterizer(render_context, filter_agents_threshold, history_num_frames,
                             render_ego_history=render_ego_history)
    elif map_type == "stub_debug":
        return StubRasterizer(render_context)
    else:
        raise NotImplementedError(f"Rasterizer for map type {map_type} is not supported.")



from typing import Tuple

import numpy as np
from shapely.geometry import LineString, Polygon


# TODO(perone):  The functions _ego_agent_within_range, _get_bounding_box,
# _get_sides would ideally be moved to an abstraction of the Agent
# later. Such as in the example below:
#
#  ego = Agent(...)
#  agent = Agent(...)
#  bbox = ego.get_bounding_box()
#  within_range = ego.within_range(agent)
#  sides = ego.get_sides()


def _get_bounding_box(centroid: np.ndarray, yaw: np.ndarray,
                      extent: np.ndarray,) -> Polygon:
    """This function will get a shapely Polygon representing the bounding box
    with an optional buffer around it.

    :param centroid: centroid of the agent
    :param yaw: the yaw of the agent
    :param extent: the extent of the agent
    :return: a shapely Polygon
    """
    x, y = centroid[0], centroid[1]
    sin, cos = np.sin(yaw), np.cos(yaw)
    width, length = extent[0] / 2, extent[1] / 2

    x1, y1 = (x + width * cos - length * sin, y + width * sin + length * cos)
    x2, y2 = (x + width * cos + length * sin, y + width * sin - length * cos)
    x3, y3 = (x - width * cos + length * sin, y - width * sin - length * cos)
    x4, y4 = (x - width * cos - length * sin, y - width * sin + length * cos)
    return Polygon([[x1, y1], [x2, y2], [x3, y3], [x4, y4]])


# TODO(perone): this should probably return a namedtuple as otherwise it
# would have to depend on the correct ordering of the front/rear/left/right
def _get_sides(bbox: Polygon) -> Tuple[LineString, LineString, LineString, LineString]:
    """This function will get the sides of a bounding box.

    :param bbox: the bounding box
    :return: a tuple with the four sides of the bounding box as LineString,
             representing front/rear/left/right.
    """
    (x1, y1), (x2, y2), (x3, y3), (x4, y4) = bbox.exterior.coords[:-1]
    return (
        LineString([(x1, y1), (x2, y2)]),
        LineString([(x3, y3), (x4, y4)]),
        LineString([(x1, y1), (x4, y4)]),
        LineString([(x2, y2), (x3, y3)]),
    )


def within_range(ego_centroid: np.ndarray, ego_extent: np.ndarray,
                 agent_centroid: np.ndarray, agent_extent: np.ndarray) -> np.ndarray:
    """This function will check if the agent is within range of the ego. It accepts
    as input a vectorized form with shapes N,D or a flat vector as well with shapes just D.

    :param ego_centroid: the ego centroid (shape: 2)
    :param ego_extent: the ego extent (shape: 3)
    :param agent_centroid: the agent centroid (shape: N, 2)
    :param agent_extent: the agent extent (shape: N, 3)
    :return: array with True if within range, False otherwise (shape: N)
    """
    distance = np.linalg.norm(ego_centroid - agent_centroid, axis=-1)
    norm_ego = np.linalg.norm(ego_extent[:2])
    norm_agent = np.linalg.norm(agent_extent[:, :2], axis=-1)
    max_range = 0.5 * (norm_ego + norm_agent)
    return distance < max_range



import warnings
from pathlib import Path

import numpy as np
import zarr
from prettytable import PrettyTable

from .filter import get_agents_slice_from_frames, get_frames_slice_from_scenes, get_tl_faces_slice_from_frames
from .labels import PERCEPTION_LABELS, TL_FACE_LABELS


# When changing the schema bump this number
FORMAT_VERSION = 2

FRAME_ARRAY_KEY = "frames"
AGENT_ARRAY_KEY = "agents"
SCENE_ARRAY_KEY = "scenes"
TL_FACE_ARRAY_KEY = "traffic_light_faces"

FRAME_CHUNK_SIZE = (10_000,)
AGENT_CHUNK_SIZE = (20_000,)
SCENE_CHUNK_SIZE = (10_000,)
TL_FACE_CHUNK_SIZE = (10_000,)

SCENE_DTYPE = [
    ("frame_index_interval", np.int64, (2,)),
    ("host", "<U16"),  # Unicode string up to 16 chars
    ("start_time", np.int64),
    ("end_time", np.int64),
]
# Information per frame (multiple per scene)
FRAME_DTYPE = [
    ("timestamp", np.int64),
    ("agent_index_interval", np.int64, (2,)),
    ("traffic_light_faces_index_interval", np.int64, (2,)),
    ("ego_translation", np.float64, (3,)),
    ("ego_rotation", np.float64, (3, 3)),
]
# Information per agent (multiple per frame)
AGENT_DTYPE = [
    ("centroid", np.float64, (2,)),
    ("extent", np.float32, (3,)),
    ("yaw", np.float32),
    ("velocity", np.float32, (2,)),
    ("track_id", np.uint64),
    ("label_probabilities", np.float32, (len(PERCEPTION_LABELS),)),
]

TL_FACE_DTYPE = [
    ("face_id", "<U16"),
    ("traffic_light_id", "<U16"),
    ("traffic_light_face_status", np.float32, (len(TL_FACE_LABELS, ))),
]


class ChunkedDataset:
    """ChunkedDataset is a dataset that lives on disk in compressed chunks, it has easy to use data loading and
    writing interfaces that involves making numpy-like slices.
    Currently only .zarr directory stores are supported (i.e. the data will live in a folder on your
    local filesystem called <something>.zarr).
    """

    def __init__(self, path: str, key: str = ""):
        """Creates a new handle for the dataset, does NOT initialize or open it yet, use respective methods for that.
        Right now only DirectoryStore is supported.

        :param path: Path on disk where to write this dataset, should end in ``.zarr``.
        :param key: Key in the zarr group to write under, you probably never need to change this (default: {""})
        """
        self.key = key
        self.path = path
        self.frames = np.empty(0, dtype=FRAME_DTYPE)
        self.scenes = np.empty(0, dtype=SCENE_DTYPE)
        self.agents = np.empty(0, dtype=AGENT_DTYPE)
        self.tl_faces = np.empty(0, dtype=TL_FACE_DTYPE)

        # Note: we still support only zarr. However, some functions build a new dataset so we cannot raise error.
        if ".zarr" not in self.path:
            warnings.warn(
                "zarr dataset path should end with .zarr (for now). Open will fail for this dataset!",
                RuntimeWarning,
                stacklevel=2,
            )
        if not Path(self.path).exists():
            warnings.warn(
                "zarr dataset path doesn't exist. Open will fail for this dataset!", RuntimeWarning, stacklevel=2
            )

    def initialize(
            self, mode: str = "w", num_scenes: int = 0, num_frames: int = 0, num_agents: int = 0, num_tl_faces: int = 0
    ) -> "ChunkedDataset":
        """Initializes a new zarr dataset, creating the underlying arrays.

        :param mode: Mode to open dataset in, should be something that supports writing. (default: {"w"})
        :param num_scenes: pre-allocate this number of scenes
        :param num_frames: pre-allocate this number of frames
        :param num_agents: pre-allocate this number of agents
        :param num_tl_faces: pre-allocate this number of traffic lights
        """

        self.root = zarr.open_group(self.path, mode=mode)

        self.frames = self.root.require_dataset(
            FRAME_ARRAY_KEY, dtype=FRAME_DTYPE, chunks=FRAME_CHUNK_SIZE, shape=(num_frames,)
        )
        self.agents = self.root.require_dataset(
            AGENT_ARRAY_KEY, dtype=AGENT_DTYPE, chunks=AGENT_CHUNK_SIZE, shape=(num_agents,)
        )
        self.scenes = self.root.require_dataset(
            SCENE_ARRAY_KEY, dtype=SCENE_DTYPE, chunks=SCENE_CHUNK_SIZE, shape=(num_scenes,)
        )
        self.tl_faces = self.root.require_dataset(
            TL_FACE_ARRAY_KEY, dtype=TL_FACE_DTYPE, chunks=TL_FACE_CHUNK_SIZE, shape=(num_tl_faces,)
        )

        self.root.attrs["format_version"] = FORMAT_VERSION
        self.root.attrs["labels"] = PERCEPTION_LABELS
        return self

    def open(self, mode: str = "r", cached: bool = True, cache_size_bytes: int = int(1e9)) -> "ChunkedDataset":
        """Opens a zarr dataset from disk from the path supplied in the constructor.

        :param mode: Mode to open dataset in, default to read-only (default: {"r"})
        :param cached: Whether to cache files read from disk using a LRU cache. (default: {True})
        :param cache_size_bytes: Size of cache in bytes (default: {1e9} (1GB))
        """
        if cached:
            self.root = zarr.open_group(
                store=zarr.LRUStoreCache(zarr.DirectoryStore(self.path), max_size=cache_size_bytes), mode=mode
            )
        else:
            self.root = zarr.open_group(self.path, mode=mode)
        self.frames = self.root[FRAME_ARRAY_KEY]
        self.agents = self.root[AGENT_ARRAY_KEY]
        self.scenes = self.root[SCENE_ARRAY_KEY]
        try:
            self.tl_faces = self.root[TL_FACE_ARRAY_KEY]
        except KeyError:
            # the real issue here is that frame doesn't have traffic_light_faces_index_interval
            warnings.warn(
                f"{TL_FACE_ARRAY_KEY} not found in {self.path}! "
                f"You won't be able to use this zarr into an Ego/AgentDataset",
                RuntimeWarning,
                stacklevel=2,
            )
            self.tl_faces = np.empty((0,), dtype=TL_FACE_DTYPE)
        return self

    def __str__(self) -> str:
        # TODO add traffic faces
        fields = [
            "Num Scenes",
            "Num Frames",
            "Num Agents",
            "Num TR lights",
            "Total Time (hr)",
            "Avg Frames per Scene",
            "Avg Agents per Frame",
            "Avg Scene Time (sec)",
            "Avg Frame frequency",
        ]
        if len(self.frames) > 1:
            # read a small chunk of frames to speed things up
            times = np.diff(self.frames[:50]["timestamp"])
            frequency = np.mean(1 / (times / 1e9))  # from nano to sec
        else:
            warnings.warn(
                f"not enough frames({len(self.frames)}) to read the frequency, 10 will be set",
                RuntimeWarning,
                stacklevel=2,
            )
            frequency = 10

        values = [
            len(self.scenes),
            len(self.frames),
            len(self.agents),
            len(self.tl_faces),
            len(self.frames) / max(frequency, 1) / 3600,
            len(self.frames) / max(len(self.scenes), 1),
            len(self.agents) / max(len(self.frames), 1),
            len(self.frames) / max(len(self.scenes), 1) / frequency,
            frequency,
        ]
        table = PrettyTable(field_names=fields)
        table.float_format = ".2"
        table.add_row(values)
        return str(table)

    def get_scene_dataset(self, scene_index: int) -> "ChunkedDataset":
        """Get a new ChunkedDataset of a single scene.
        This dataset lives in memory (as np.ndarray)

        :param scene_index: the scene index
        :return: a dataset with a single scene inside
        """
        if scene_index >= len(self.scenes):
            raise ValueError(f"scene index {scene_index} out of bound for dataset with {len(self.scenes)} scenes")

        scenes = self.scenes[scene_index: scene_index + 1].copy()
        frame_slice = get_frames_slice_from_scenes(*scenes)
        frames = self.frames[frame_slice].copy()
        agent_slice = get_agents_slice_from_frames(*frames[[0, -1]])
        tl_slice = get_tl_faces_slice_from_frames(*frames[[0, -1]])

        agents = self.agents[agent_slice].copy()
        tl_faces = self.tl_faces[tl_slice].copy()

        frames["agent_index_interval"] -= agent_slice.start
        frames["traffic_light_faces_index_interval"] -= tl_slice.start
        scenes["frame_index_interval"] -= frame_slice.start

        dataset = ChunkedDataset("")
        dataset.agents = agents
        dataset.tl_faces = tl_faces
        dataset.frames = frames
        dataset.scenes = scenes

        return dataset



from collections import defaultdict, OrderedDict
from typing import List

import numpy as np

from .csv_utils import read_gt_csv, read_pred_csv
from .metrics import metric_signature


def validate_dicts(ground_truth: dict, predicted: dict) -> bool:
    """
    Validate GT and pred dictionaries by comparing keys

    Args:
        ground_truth (dict): mapping from (track_id + timestamp) to an element returned from our csv utils
        predicted (dict): mapping from (track_id + timestamp) to an element returned from our csv utils

    Returns:
        (bool): True if the 2 dicts match (same keys)

    """
    valid = True

    num_agents_gt = len(ground_truth)
    num_agents_pred = len(predicted)

    if num_agents_gt != num_agents_pred:
        print(f"Incorrect number of rows in inference csv. Expected {num_agents_gt}, Got {num_agents_pred}")
        valid = False

    missing_agents = ground_truth.keys() - predicted.keys()
    if len(missing_agents):
        valid = False

    for missing_agents in missing_agents:
        print(f"Missing agents: {missing_agents}")

    unknown_agents = predicted.keys() - ground_truth.keys()
    if len(unknown_agents):
        valid = False

    for unknown_agent in unknown_agents:
        print(f"Unknown agents: {unknown_agent}")

    return valid


def compute_metrics_csv(ground_truth_path: str, inference_output_path: str, metrics: List[metric_signature]) -> dict:
    """
    Compute a set of metrics between ground truth and prediction csv files

    Arguments:
        ground_truth_path (str): Path to the ground truth csv file.
        inference_output_path (str): Path to the csv file containing network output.
        metrics (List[Callable]): a list of callable to be applied to the elements retrieved from the 2
        csv files

    Returns:
        dict: keys are metrics name, values is the average metric computed over the elements
    """

    assert len(metrics) > 0, "you must pass at least one metric to compute"

    ground_truth = OrderedDict()
    inference = OrderedDict()

    for el in read_gt_csv(ground_truth_path):
        ground_truth[el["track_id"] + el["timestamp"]] = el
    for el in read_pred_csv(inference_output_path):
        inference[el["track_id"] + el["timestamp"]] = el

    if not validate_dicts(ground_truth, inference):
        raise ValueError("Error validating csv, see above for details.")

    metrics_dict = defaultdict(list)

    for key, ground_truth_value in ground_truth.items():
        gt_coord = ground_truth_value["coord"]
        avail = ground_truth_value["avail"]

        pred_coords = inference[key]["coords"]
        conf = inference[key]["conf"]
        for metric in metrics:
            metrics_dict[metric.__name__].append(metric(gt_coord, pred_coords, conf, avail))

    # compute average of each metric
    return {metric_name: np.mean(values, axis=0) for metric_name, values in metrics_dict.items()}



import math
from abc import ABC, abstractmethod
from typing import Dict

import numpy as np

from l5kit.simulation.unroll import TrajectoryStateIndices


class KinematicModel(ABC):
    """Base class interface for kinematic model."""
    #: The prefix that will identify the model class
    model_prefix: str

    @abstractmethod
    def reset(self, init_state: np.ndarray) -> None:
        """Reset the model state when new episode starts.

        :param init_state: the initial state of the ego
        """
        raise NotImplementedError

    @abstractmethod
    def update(self, input_action: TrajectoryStateIndices) -> Dict[str, np.ndarray]:
        """Update the model state based on the action at a particular time-step during the episode.

        :param input_action: the prediction for next time-step
        :return: reward at a particular frame index (time-step) during the episode.
        """
        raise NotImplementedError


class UnicycleModel(KinematicModel):
    """This class is responsible for controlling kinematics using the Unicycle model.

    :param model_prefix: the prefix that will identify this model class
    :param min_acc: the threshold for minimum acceleration
    :param max_acc: the threshold for maximum acceleration
    :param min_steer: the threshold for minimum steering
    :param max_steer: the threshold for maximum steering
    """

    def __init__(self, model_prefix: str = "Unicycle",
                 min_acc: float = -0.6,  # min acceleration: -6 mps2
                 max_acc: float = 0.6,   # max acceleration: 6 mps2
                 min_steer: float = -math.radians(45) * 0.1,  # max yaw rate: 45 degrees per second
                 max_steer: float = math.radians(45) * 0.1,   # max yaw rate: 45 degrees per second
                 ) -> None:
        """Constructor method
        """
        self.model_prefix = model_prefix
        # Thresholds
        self.min_acc = min_acc
        self.max_acc = max_acc
        self.min_steer = min_steer
        self.max_steer = max_steer

    def reset(self, init_state: np.ndarray) -> None:
        """Reset the model state when new episode starts.

        :param init_state: the initial state of the ego
        """
        self.old_x = init_state[TrajectoryStateIndices.X]
        self.old_y = init_state[TrajectoryStateIndices.Y]
        self.old_r = init_state[TrajectoryStateIndices.THETA]
        self.old_v = init_state[TrajectoryStateIndices.SPEED]

    def update(self, input_action: np.ndarray) -> Dict[str, np.ndarray]:
        """Update the model state based on the action at a particular time-step during the episode.

        :param input_action: the prediction [steer, acc] for next time-step
        :return: reward at a particular frame index (time-step) during the episode.
        """

        steer = input_action[0]
        acc = input_action[1]

        # Clip
        steer = np.clip(steer, self.min_steer, self.max_steer)
        acc = np.clip(acc, self.min_acc, self.max_acc)

        # Update x, y, r, v
        # x, y and r are already zero-centered in the raster image
        self.new_r = steer
        self.new_v = self.old_v + acc
        self.new_x = math.cos(self.new_r) * self.new_v
        self.new_y = math.sin(self.new_r) * self.new_v

        # Predicted positions and yaws according to unicycle model
        return_dict = {"positions": np.array([[[self.new_x, self.new_y]]]), "yaws": np.array([[[self.new_r]]])}

        # Saved updated states
        self.old_v = self.new_v

        return return_dict



import argparse
from pathlib import Path

import numpy as np
from zarr import convenience

from l5kit.data import ChunkedDataset, get_agents_slice_from_frames
from l5kit.data.zarr_utils import zarr_scenes_chop
from l5kit.dataset.select_agents import select_agents, TH_DISTANCE_AV, TH_EXTENT_RATIO, TH_YAW_DEGREE

from .extract_ground_truth import export_zarr_to_csv


MIN_FUTURE_STEPS = 10


def create_chopped_dataset(
        zarr_path: str, th_agent_prob: float, num_frames_to_copy: int, num_frames_gt: int, min_frame_future: int
) -> str:
    """
    Create a chopped version of the zarr that can be used as a test set.
    This function was used to generate the test set for the competition so that the future GT is not in the data.

    Store:
     - a dataset where each scene has been chopped at `num_frames_to_copy` frames;
     - a mask for agents for those final frames based on the original mask and a threshold on the future_frames;
     - the GT csv for those agents

     For the competition, only the first two (dataset and mask) will be available in the notebooks

    Args:
        zarr_path (str): input zarr path to be chopped
        th_agent_prob (float): threshold over agents probabilities used in select_agents function
        num_frames_to_copy (int):  number of frames to copy from the beginning of each scene, others will be discarded
        min_frame_future (int): minimum number of frames that must be available in the future for an agent
        num_frames_gt (int): number of future predictions to store in the GT file

    Returns:
        str: the parent folder of the new datam
    """
    zarr_path = Path(zarr_path)
    dest_path = zarr_path.parent / f"{zarr_path.stem}_chopped_{num_frames_to_copy}"
    chopped_path = dest_path / zarr_path.name
    gt_path = dest_path / "gt.csv"
    mask_chopped_path = dest_path / "mask"

    # Create standard mask for the dataset so we can use it to filter out unreliable agents
    zarr_dt = ChunkedDataset(str(zarr_path))
    zarr_dt.open()

    agents_mask_path = Path(zarr_path) / f"agents_mask/{th_agent_prob}"
    if not agents_mask_path.exists():  # don't check in root but check for the path
        select_agents(
            zarr_dt,
            th_agent_prob=th_agent_prob,
            th_yaw_degree=TH_YAW_DEGREE,
            th_extent_ratio=TH_EXTENT_RATIO,
            th_distance_av=TH_DISTANCE_AV,
        )
    agents_mask_origin = np.asarray(convenience.load(str(agents_mask_path)))

    # create chopped dataset
    zarr_scenes_chop(str(zarr_path), str(chopped_path), num_frames_to_copy=num_frames_to_copy)
    zarr_chopped = ChunkedDataset(str(chopped_path))
    zarr_chopped.open()

    # compute the chopped boolean mask, but also the original one limited to frames of interest for GT csv
    agents_mask_chop_bool = np.zeros(len(zarr_chopped.agents), dtype=np.bool)
    agents_mask_orig_bool = np.zeros(len(zarr_dt.agents), dtype=np.bool)

    for idx in range(len(zarr_dt.scenes)):
        scene = zarr_dt.scenes[idx]

        frame_original = zarr_dt.frames[scene["frame_index_interval"][0] + num_frames_to_copy - 1]
        slice_agents_original = get_agents_slice_from_frames(frame_original)
        frame_chopped = zarr_chopped.frames[zarr_chopped.scenes[idx]["frame_index_interval"][-1] - 1]
        slice_agents_chopped = get_agents_slice_from_frames(frame_chopped)

        mask = agents_mask_origin[slice_agents_original][:, 1] >= min_frame_future
        agents_mask_orig_bool[slice_agents_original] = mask.copy()
        agents_mask_chop_bool[slice_agents_chopped] = mask.copy()

    # store the mask and the GT csv of frames on interest
    np.savez(str(mask_chopped_path), agents_mask_chop_bool)
    export_zarr_to_csv(zarr_dt, str(gt_path), num_frames_gt, th_agent_prob, agents_mask=agents_mask_orig_bool)
    return str(dest_path)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--zarr_paths", required=True, nargs="+", type=Path)
    parser.add_argument("--th_agent_prob", type=float, required=True, help="perception threshold on agents")
    parser.add_argument("--num_frames_to_copy", required=True, type=int)
    parser.add_argument("--future_steps", required=True, type=int)
    parser.add_argument("--min_future_steps", default=MIN_FUTURE_STEPS, type=int)
    args = parser.parse_args()
    for zarr_path in args.zarr_paths:
        create_chopped_dataset(
            zarr_path, args.th_agent_prob, args.num_frames_to_copy, args.future_steps, args.min_future_steps
        )



import numpy as np

from l5kit.data import ChunkedDataset, get_agents_slice_from_frames, get_tl_faces_slice_from_frames


def insert_agent(agent: np.ndarray, frame_idx: int, dataset: ChunkedDataset) -> None:
    """Insert an agent in one frame.
    Assumptions:
    - the dataset has only 1 scene
    - the dataset is in numpy format and not zarr anymore

    :param agent: the agent info to be inserted
    :param frame_idx: the frame where we want to insert the agent
    :param dataset: the single-scene dataset.
    """
    if not len(dataset.scenes) == 1:
        raise ValueError(f"dataset should have a single scene, got {len(dataset.scenes)}")
    if not isinstance(dataset.agents, np.ndarray):
        raise ValueError("dataset agents should be an editable np array")
    if not isinstance(dataset.frames, np.ndarray):
        raise ValueError("dataset frames should be an editable np array")
    if not frame_idx < len(dataset.frames):
        raise ValueError(f"can't set frame {frame_idx} in dataset with len {len(dataset.frames)}")

    frame = dataset.frames[frame_idx]
    agents_slice = get_agents_slice_from_frames(frame)
    agents_frame = dataset.agents[agents_slice]

    idx_set = np.argwhere(agent["track_id"] == agents_frame["track_id"])
    assert len(idx_set) in [0, 1]

    if len(idx_set):
        # CASE 1
        # the agent is already there and we can just update it
        # we set also label_probabilities from the current one to ensure it is high enough
        idx_set = int(idx_set[0])
        agents_frame[idx_set: idx_set + 1] = agent
    else:
        # CASE 2
        # we need to insert the agent and move everything
        dataset.agents = np.concatenate(
            [dataset.agents[0: agents_slice.stop], agent, dataset.agents[agents_slice.stop:]], 0
        )

        # move end of the current frame and all other frames start and end
        dataset.frames[frame_idx]["agent_index_interval"] += (0, 1)
        dataset.frames[frame_idx + 1:]["agent_index_interval"] += 1


def disable_agents(dataset: ChunkedDataset, allowlist: np.ndarray) -> None:
    """Disable all agents in dataset except for the ones in allowlist
    Assumptions:
    - the dataset has only 1 scene
    - the dataset is in numpy format and not zarr anymore

    :param dataset: the single-scene dataset
    :param allowlist: 1D np array of track_ids to keep

    """
    if not len(dataset.scenes) == 1:
        raise ValueError(f"dataset should have a single scene, got {len(dataset.scenes)}")
    if not isinstance(dataset.agents, np.ndarray):
        raise ValueError("dataset agents should be an editable np array")
    if not len(allowlist.shape) == 1:
        raise ValueError("allow list should be 1D")

    agent_track_ids = dataset.agents["track_id"]

    mask_disable = ~np.in1d(agent_track_ids, allowlist)

    # this will set those agents as invisible
    # we also zeroes their pose and extent
    dataset.agents["centroid"][mask_disable] *= 0
    dataset.agents["yaw"][mask_disable] *= 0
    dataset.agents["extent"][mask_disable] *= 0
    dataset.agents["label_probabilities"][mask_disable] = -1


def get_frames_subset(dataset: ChunkedDataset, frame_start_idx: int, frame_end_idx: int) -> ChunkedDataset:
    """Get a new dataset with frames between start (included) and end (excluded).
    Assumptions:
    - the dataset has only 1 scene
    - the dataset is in numpy format and not zarr anymore

    :param dataset: the single-scene dataset.
    :param frame_start_idx: first frame to keep.
    :param frame_end_idx: where to stop taking frames (excluded).

    """
    if not len(dataset.scenes) == 1:
        raise ValueError(f"dataset should have a single scene, got {len(dataset.scenes)}")
    if not isinstance(dataset.agents, np.ndarray):
        raise ValueError("dataset agents should be an editable np array")
    if not isinstance(dataset.tl_faces, np.ndarray):
        raise ValueError("dataset tls should be an editable np array")
    if not isinstance(dataset.frames, np.ndarray):
        raise ValueError("dataset frames should be an editable np array")
    if frame_start_idx >= len(dataset.frames):
        raise ValueError(f"frame start {frame_start_idx} is over the length of the dataset")
    if frame_end_idx > len(dataset.frames):
        raise ValueError(f"frame end {frame_end_idx} is over the length of the dataset")
    if frame_start_idx >= frame_end_idx:
        raise ValueError(f"end frame {frame_end_idx} should be higher than start {frame_start_idx}")
    if frame_start_idx < 0:
        raise ValueError(f"start frame {frame_start_idx} should be positive")

    new_dataset = ChunkedDataset("")
    new_dataset.scenes = dataset.scenes.copy()
    new_dataset.scenes[0]["start_time"] = dataset.frames[frame_start_idx]["timestamp"]
    new_dataset.scenes[0]["end_time"] = dataset.frames[frame_end_idx - 1]["timestamp"]

    new_dataset.frames = dataset.frames[frame_start_idx:frame_end_idx].copy()
    new_dataset.scenes[0]["frame_index_interval"] = (0, len(new_dataset.frames))

    agent_slice = get_agents_slice_from_frames(*dataset.frames[[frame_start_idx, frame_end_idx - 1]])
    tls_slice = get_tl_faces_slice_from_frames(*dataset.frames[[frame_start_idx, frame_end_idx - 1]])
    new_dataset.frames["agent_index_interval"] -= new_dataset.frames["agent_index_interval"][0, 0]
    new_dataset.frames["traffic_light_faces_index_interval"] -= new_dataset.frames[
        "traffic_light_faces_index_interval"
    ][0, 0]
    new_dataset.agents = dataset.agents[agent_slice].copy()
    new_dataset.tl_faces = dataset.tl_faces[tls_slice].copy()
    return new_dataset



from .angle import angle_between_vectors, angular_distance, compute_yaw_around_north_from_direction
from .image import crop_rectangle_from_image
from .transform import (compute_agent_pose, ecef_to_geodetic, geodetic_to_ecef, rotation33_as_yaw, transform_point,
                        transform_points, vertical_flip, yaw_as_rotation33)
from .voxel import normalize_intensity, points_within_bounds, voxel_coords_to_intensity_grid


__all__ = [
    "angle_between_vectors",
    "compute_yaw_around_north_from_direction",
    "crop_rectangle_from_image",
    "rotation33_as_yaw",
    "yaw_as_rotation33",
    "vertical_flip",
    "transform_points",
    "transform_point",
    "ecef_to_geodetic",
    "geodetic_to_ecef",
    "points_within_bounds",
    "voxel_coords_to_intensity_grid",
    "normalize_intensity",
    "angular_distance",
    "compute_agent_pose",
]



import csv
from itertools import chain
from typing import Iterator, List, Optional

import numpy as np


MAX_MODES = 3

"""
These utilities write and read csv with ground-truth and prediction data.
Both share the first two fields (timestamp and track_id) which are used to identify a single record.

GT:
Single mode future prediction with availabilities (either 1->available or 0->unavailable).
Header fields have these meanings:
timestamp, track_id, avail_time_0, avail_time_1, ..., coord_x_time_0, coord_y_time_0, ...

PRED:
Multi mode future prediction with confidence score (one value per mode up to MAX_MODES, sum to 1).
Header fields have these meanings:
timestamp, track_id, conf_mode_0, conf_mode_1, ..., coord_x_time_0_mode_0, coord_y_time_0_mode_0, ...,
coord_x_time_0_mode_1, coord_y_time_0_mode_1, ...

"""


def _generate_coords_keys(future_len: int, mode_index: int = 0) -> List[str]:
    """
    Generate keys like coord_x00, coord_y00... that can be used to get or set value in CSV.
    Two keys for each mode and future step.

    Args:
        future_len (int): how many prediction the data has in the future
        mode_index (int): what mode are we reading/writing

    Returns:
        List[str]: a list of keys
    """
    return list(
        chain.from_iterable([[f"coord_x{mode_index}{i}", f"coord_y{mode_index}{i}"] for i in range(future_len)])
    )


def _generate_avails_keys(future_len: int) -> List[str]:
    """
    Generate availabilities keys (one per future step)

    Args:
        future_len (int): how many prediction in the future data has

    Returns:
        List[str]: a list of keys
    """
    return [f"avail_{i}" for i in range(future_len)]


def _generate_confs_keys() -> List[str]:
    """
    Generate modes keys (one per mode)

    Returns:
        List[str]: a list of keys
    """
    return [f"conf_{i}" for i in range(MAX_MODES)]


def write_gt_csv(
        csv_path: str, timestamps: np.ndarray, track_ids: np.ndarray, coords: np.ndarray, avails: np.ndarray
) -> None:
    """
    Encode the ground truth into a csv file

    Args:
        csv_path (str): path to the csv to write
        timestamps (np.ndarray): (num_example,) frame timestamps
        track_ids (np.ndarray): (num_example,) agent ids
        coords (np.ndarray): (num_example x future_len x num_coords) meters displacements
        avails (np.ndarray): (num_example x future_len) array with value 0 (discard in eval) or 1 (keep in eval)

    Returns:

    """
    assert len(coords.shape) == 3
    num_example, future_len, num_coords = coords.shape
    assert num_coords == 2
    assert timestamps.shape == track_ids.shape == (num_example,)
    assert avails.shape == (num_example, future_len)

    coords_keys = _generate_coords_keys(future_len)
    avails_keys = _generate_avails_keys(future_len)

    # create and write HEADER
    # order is (timestamp,track_id,availabilities,coords)
    fieldnames = ["timestamp", "track_id"] + avails_keys + coords_keys
    writer = csv.DictWriter(open(csv_path, "w"), fieldnames)
    writer.writeheader()

    for timestamp, track_id, coord, avail in zip(timestamps, track_ids, coords, avails):
        # writing using dict relieves us from respecting the order
        line = {"timestamp": timestamp, "track_id": track_id}
        line.update({key: ava for key, ava in zip(avails_keys, avail)})
        line.update({key: f"{cor:.5f}" for key, cor in zip(coords_keys, coord.reshape(-1))})

        writer.writerow(line)


def read_gt_csv(csv_path: str) -> Iterator[dict]:
    """
    Generator function that returns a line at a time from the csv file as a dict

    Args:
        csv_path (str): path of the csv to read

    Returns:
        Iterator[dict]: dict keys are the csv header fieldnames
    """
    reader = csv.DictReader(open(csv_path, "r"))
    fieldnames = reader.fieldnames
    assert fieldnames is not None, "error reading fieldnames"

    future_len = (len(fieldnames) - 2) / 3  # excluded first two fields, the rest must be (x, y, av) * len = 3*len
    assert future_len == int(future_len), "error estimating len"
    future_len = int(future_len)

    coords_keys = _generate_coords_keys(future_len)
    avails_keys = _generate_avails_keys(future_len)

    for row in reader:
        timestamp = row["timestamp"]
        track_id = row["track_id"]

        avail = np.asarray([np.float64(row[key]) for key in avails_keys])

        coord = np.asarray([np.float64(row[key]) for key in coords_keys])
        coord = coord.reshape((future_len, 2))

        yield {"track_id": track_id, "timestamp": timestamp, "coord": coord, "avail": avail}


def write_pred_csv(
        csv_path: str,
        timestamps: np.ndarray,
        track_ids: np.ndarray,
        coords: np.ndarray,
        confs: Optional[np.ndarray] = None,
) -> None:
    """
    Encode the predictions into a csv file. Coords can have an additional axis for multi-mode.
    We handle up to MAX_MODES modes.
    For the uni-modal case (i.e. all predictions have just a single mode), coords should not have the additional axis
    and confs should be set to None. In this case, a single mode with confidence 1 will be written.

    Args:
        csv_path (str): path to the csv to write
        timestamps (np.ndarray): (num_example,) frame timestamps
        track_ids (np.ndarray): (num_example,) agent ids
        coords (np.ndarray): (num_example x (modes) x future_len x num_coords) meters displacements
        confs (Optional[np.ndarray]): (num_example x modes) confidence of each modes in each example.
        Rows should sum to 1

    Returns:

    """
    assert len(coords.shape) in [3, 4]

    if len(coords.shape) == 3:
        assert confs is None  # no conf for the single-mode case
        coords = np.expand_dims(coords, 1)  # add a new axis for the multi-mode
        confs = np.ones((len(coords), 1))  # full confidence

    num_example, num_modes, future_len, num_coords = coords.shape
    assert num_coords == 2
    assert timestamps.shape == track_ids.shape == (num_example,)
    assert confs is not None and confs.shape == (num_example, num_modes)
    assert np.allclose(np.sum(confs, axis=-1), 1.0)
    assert num_modes <= MAX_MODES

    # generate always a fixed size json for MAX_MODES by padding the arrays with zeros
    coords_padded = np.zeros((num_example, MAX_MODES, future_len, num_coords), dtype=coords.dtype)
    coords_padded[:, :num_modes] = coords
    confs_padded = np.zeros((num_example, MAX_MODES), dtype=confs.dtype)
    confs_padded[:, :num_modes] = confs

    coords_keys_list = [_generate_coords_keys(future_len, mode_index=idx) for idx in range(MAX_MODES)]
    confs_keys = _generate_confs_keys()

    # create and write HEADER
    # order is (timestamp,track_id,confidences,coords)
    fieldnames = ["timestamp", "track_id"] + confs_keys  # all confidences before coordinates
    for coords_labels in coords_keys_list:
        fieldnames.extend(coords_labels)

    writer = csv.DictWriter(open(csv_path, "w"), fieldnames)
    writer.writeheader()

    for timestamp, track_id, coord, conf in zip(timestamps, track_ids, coords_padded, confs_padded):
        line = {"timestamp": timestamp, "track_id": track_id}
        line.update({key: con for key, con in zip(confs_keys, conf)})

        for idx in range(MAX_MODES):
            line.update({key: f"{cor:.5f}" for key, cor in zip(coords_keys_list[idx], coord[idx].reshape(-1))})

        writer.writerow(line)


def read_pred_csv(csv_path: str) -> Iterator[dict]:
    """
    Generator function that returns a line at the time from the csv file as a dict

    Args:
        csv_path (str): path of the csv to read

    Returns:
        Iterator[dict]: dict keys are the csv header fieldnames
    """

    reader = csv.DictReader(open(csv_path, "r"))
    fieldnames = reader.fieldnames
    assert fieldnames is not None, "error reading fieldnames"

    # exclude timestamp, track_id and MAX_MODES confs, the rest should be (x, y) * len * 3 = 6*len
    future_len = (len(fieldnames) - (2 + MAX_MODES)) / 6
    assert future_len == int(future_len), "error estimating len"
    future_len = int(future_len)

    coords_labels_list = [_generate_coords_keys(future_len, mode_index=idx) for idx in range(MAX_MODES)]
    confs_labels = _generate_confs_keys()

    for row in reader:
        track_id = row["track_id"]
        timestamp = row["timestamp"]

        conf = np.asarray([np.float64(row[conf_label]) for conf_label in confs_labels])

        coords = []
        for idx in range(MAX_MODES):
            coord = np.asarray([np.float64(row[coord_label]) for coord_label in coords_labels_list[idx]])
            coords.append(coord.reshape((future_len, 2)))

        coords = np.stack(coords, axis=0)

        yield {"track_id": track_id, "timestamp": timestamp, "coords": coords, "conf": conf}



from typing import List, Optional, Tuple, Union

import cv2
import numpy as np

from l5kit.data.zarr_dataset import AGENT_DTYPE

from ..data.filter import filter_agents_by_labels, filter_agents_by_track_id
from ..geometry import rotation33_as_yaw, transform_points
from .rasterizer import EGO_EXTENT_HEIGHT, EGO_EXTENT_LENGTH, EGO_EXTENT_WIDTH, Rasterizer
from .render_context import RenderContext
from .semantic_rasterizer import CV2_SUB_VALUES, cv2_subpixel


# TODO this can be useful to have around
def get_ego_as_agent(frame: np.ndarray) -> np.ndarray:
    """Get a valid agent with information from the AV. Ford Fusion extent is used.

    :param frame: The frame from which the Ego states are extracted
    :return: An agent numpy array of the Ego states
    """
    ego_agent = np.zeros(1, dtype=AGENT_DTYPE)
    ego_agent[0]["centroid"] = frame["ego_translation"][:2]
    ego_agent[0]["yaw"] = rotation33_as_yaw(frame["ego_rotation"])
    ego_agent[0]["extent"] = np.asarray((EGO_EXTENT_LENGTH, EGO_EXTENT_WIDTH, EGO_EXTENT_HEIGHT))
    return ego_agent


def get_box_world_coords(agents: np.ndarray) -> np.ndarray:
    """Get world coordinates of the 4 corners of the bounding boxes

    :param agents: agents array of size N with centroid (world coord), yaw and extent
    :return: array of shape (N, 4, 2) with the four corners of each agent
    """
    # shape is (1, 4, 2)
    corners_base_coords = (np.asarray([[-1, -1], [-1, 1], [1, 1], [1, -1]]) * 0.5)[None, :, :]

    # compute the corner in world-space (start in origin, rotate and then translate)
    # extend extent to shape (N, 1, 2) so that op gives (N, 4, 2)
    corners_m = corners_base_coords * agents["extent"][:, None, :2]  # corners in zero
    s = np.sin(agents["yaw"])
    c = np.cos(agents["yaw"])
    # note this is clockwise because it's right-multiplied and not left-multiplied later,
    # and therefore we're still rotating counterclockwise.
    rotation_m = np.moveaxis(np.array(((c, s), (-s, c))), 2, 0)
    # extend centroid to shape (N, 1, 2) so that op gives (N, 4, 2)
    box_world_coords = corners_m @ rotation_m + agents["centroid"][:, None, :2]
    return box_world_coords


def draw_boxes(
        raster_size: Tuple[int, int],
        raster_from_world: np.ndarray,
        agents: np.ndarray,
        color: Union[int, Tuple[int, int, int]],
) -> np.ndarray:
    """ Draw multiple boxes in one sweep over the image.
    Boxes corners are extracted from agents, and the coordinates are projected on the image space.
    Finally, cv2 draws the boxes.

    :param raster_size: Desired output raster image size
    :param raster_from_world: Transformation matrix to transform from world to image coordinate
    :param agents: An array of agents to be drawn
    :param color: Single int or RGB color
    :return: the image with agents rendered. A RGB image if using RGB color, otherwise a GRAY image
    """
    if isinstance(color, int):
        im = np.zeros((raster_size[1], raster_size[0]), dtype=np.uint8)
    else:
        im = np.zeros((raster_size[1], raster_size[0], 3), dtype=np.uint8)

    box_world_coords = get_box_world_coords(agents)
    box_raster_coords = transform_points(box_world_coords.reshape((-1, 2)), raster_from_world)

    # fillPoly wants polys in a sequence with points inside as (x,y)
    box_raster_coords = cv2_subpixel(box_raster_coords.reshape((-1, 4, 2)))
    cv2.fillPoly(im, box_raster_coords, color=color, **CV2_SUB_VALUES)
    return im


class BoxRasterizer(Rasterizer):
    def __init__(
        self,
        render_context: RenderContext,
        filter_agents_threshold: float,
        history_num_frames: int,
        render_ego_history: bool = True,
    ) -> None:
        """This is a rasterizer class used for rendering agents' bounding boxes on the raster image.

        :param render_context: Render context
        :param filter_agents_threshold: Value between 0 and 1 used to filter uncertain agent detections
        :param history_num_frames: Number of past frames to be renderd on the raster
        :param render_ego_history: Option to render past ego states on the raster image
        """
        super(BoxRasterizer, self).__init__()
        self.render_context = render_context
        self.raster_size = render_context.raster_size_px
        self.filter_agents_threshold = filter_agents_threshold
        self.history_num_frames = history_num_frames
        self.render_ego_history = render_ego_history

    def rasterize(
            self,
            history_frames: np.ndarray,
            history_agents: List[np.ndarray],
            history_tl_faces: List[np.ndarray],
            agent: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Generate raster image by rendering Ego & Agents as bounding boxes on the raster image.
        Ego & Agents from different past frame are rendered at different image channel.

        :param history_frames: A list of past frames to be rasterized
        :param history_agents: A list of agents from past frames to be rasterized
        :param history_tl_faces: A list of traffic light faces from past frames to be rasterized
        :param agent: The selected agent to be rendered as Ego, if it is None the AV will be rendered as Ego
        :return: An raster image of size [2xN] with Ego & Agents rendered as bounding boxes, where N is number of
         history frames
        """
        # all frames are drawn relative to this one"
        frame = history_frames[0]
        if agent is None:
            ego_translation_m = history_frames[0]["ego_translation"]
            ego_yaw_rad = rotation33_as_yaw(frame["ego_rotation"])
        else:
            ego_translation_m = np.append(agent["centroid"], history_frames[0]["ego_translation"][-1])
            ego_yaw_rad = agent["yaw"]

        raster_from_world = self.render_context.raster_from_world(ego_translation_m, ego_yaw_rad)

        # this ensures we always end up with fixed size arrays, +1 is because current time is also in the history
        out_shape = (self.raster_size[1], self.raster_size[0], self.history_num_frames + 1)
        agents_images = np.zeros(out_shape, dtype=np.uint8)
        ego_images = np.zeros(out_shape, dtype=np.uint8)

        for i, (frame, agents) in enumerate(zip(history_frames, history_agents)):
            agents = filter_agents_by_labels(agents, self.filter_agents_threshold)
            # note the cast is for legacy support of dataset before April 2020
            av_agent = get_ego_as_agent(frame).astype(agents.dtype)

            if agent is None:
                ego_agent = av_agent
            else:
                ego_agent = filter_agents_by_track_id(agents, agent["track_id"])
                agents = np.append(agents, av_agent)  # add av_agent to agents
                if len(ego_agent) > 0:  # check if ego_agent is in the frame
                    agents = agents[agents != ego_agent[0]]  # remove ego_agent from agents

            agents_images[..., i] = draw_boxes(self.raster_size, raster_from_world, agents, 255)
            if len(ego_agent) > 0 and (self.render_ego_history or i == 0):
                ego_images[..., i] = draw_boxes(self.raster_size, raster_from_world, ego_agent, 255)

        # combine such that the image consists of [agent_t, agent_t-1, agent_t-2, ego_t, ego_t-1, ego_t-2]
        out_im = np.concatenate((agents_images, ego_images), -1)

        return out_im.astype(np.float32) / 255

    def to_rgb(self, in_im: np.ndarray, **kwargs: dict) -> np.ndarray:
        """This function is used to get an rgb image where agents further in the past have faded colors.

        :param in_im: The output of the rasterize function
        :param kwargs: This can be used for additional customization (such as colors)
        :return: An RGB image with agents and ego coloured with fading colors
        """
        hist_frames = in_im.shape[-1] // 2
        in_im = np.transpose(in_im, (2, 0, 1))

        # this is similar to the draw history code
        out_im_agent = np.zeros((self.raster_size[1], self.raster_size[0], 3), dtype=np.float32)
        agent_chs = in_im[:hist_frames][::-1]  # reverse to start from the furthest one
        agent_color = (0, 0, 1) if "agent_color" not in kwargs else kwargs["agent_color"]
        for ch in agent_chs:
            out_im_agent *= 0.85  # magic fading constant for the past
            out_im_agent[ch > 0] = agent_color

        out_im_ego = np.zeros((self.raster_size[1], self.raster_size[0], 3), dtype=np.float32)
        ego_chs = in_im[hist_frames:][::-1]
        ego_color = (0, 1, 0) if "ego_color" not in kwargs else kwargs["ego_color"]
        for ch in ego_chs:
            out_im_ego *= 0.85
            out_im_ego[ch > 0] = ego_color

        out_im = (np.clip(out_im_agent + out_im_ego, 0, 1) * 255).astype(np.uint8)
        return out_im

    def num_channels(self) -> int:
        return (self.history_num_frames + 1) * 2



from typing import cast, Union

import numpy as np


def angle_between_vectors(v1: np.ndarray, v2: np.ndarray) -> float:
    """angle_between_vectors returns the angle in radians between two vectors.

    Args:
        v1 (np.ndarray): Vector 1 of shape (N)
        v2 (np.ndarray): Vector 2 of same shape as ``v1``

    Returns:
        float: angle in radians
    """
    cos_ang = np.dot(v1, v2)
    sin_ang = np.linalg.norm(np.cross(v1, v2))
    return cast(float, np.arctan2(sin_ang, cos_ang))


def compute_yaw_around_north_from_direction(direction_vector: np.ndarray) -> float:
    """compute_yaw_from_direction computes the yaw as angle between a 2D input direction vector and
the y-axis direction vector (0, 1).

    Args:
        direction_vector (np.ndarray): Vector of shape (2,)

    Returns:
        float: angle to (0,1) vector in radians
    """
    return angle_between_vectors(direction_vector, np.array([0.0, 1.0]))


def angular_distance(angle_a: Union[float, np.ndarray], angle_b: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
    """ A function that takes two arrays of angles in radian and compute the angular distance, wrap the angular
    distance such that they are always in the [-pi, pi) range.

    Args:
        angle_a (np.ndarray, float): first array of angles in radians
        angle_b (np.ndarray, float): second array of angles in radians

    Returns:
        angular distance in radians between two arrays of angles
    """

    return (angle_a - angle_b + np.pi) % (2 * np.pi) - np.pi



import os
from collections import Counter
from pathlib import Path
from typing import List, Optional, Tuple

import numpy as np
from tqdm import tqdm

from .filter import get_agents_slice_from_frames, get_frames_slice_from_scenes, get_tl_faces_slice_from_frames
from .zarr_dataset import ChunkedDataset


GIGABYTE = 1 * 1024 * 1024 * 1024


def _compute_path_size(path: str) -> int:
    """
    Compute the total size of the folder, considering also nested elements.
    Can be run to get zarr total size

    Args:
        path (str): base path

    Returns:
        (int): total size in bytes
    """
    root_directory = Path(path)
    return sum(f.stat().st_size for f in root_directory.glob("**/*") if f.is_file())


def _get_num_els_in_scene_range(zarr_dataset: ChunkedDataset, scene_index_start: int, scene_index_end: int) -> dict:
    """
    Get numbers of scenes, frames, agents, tl_lights in a set of scene in a zarr
    Args:
        zarr_dataset (ChunkedDataset): zarr dataset to use for computing number of elements
        scene_index_start (int): start from this scene (included)
        scene_index_end (int): end before this scene (excluded!!)

    Returns:
        dict: a dict with keys equal to zarr initialise args
    """
    assert scene_index_end > scene_index_start

    scene_start = zarr_dataset.scenes[scene_index_start]
    scene_end = zarr_dataset.scenes[scene_index_end - 1]

    frame_start = zarr_dataset.frames[scene_start["frame_index_interval"][0]]
    frame_end = zarr_dataset.frames[scene_end["frame_index_interval"][1] - 1]

    return {
        "num_scenes": scene_index_end - scene_index_start,
        "num_frames": scene_end["frame_index_interval"][1] - scene_start["frame_index_interval"][0],
        "num_agents": frame_end["agent_index_interval"][1] - frame_start["agent_index_interval"][0],
        "num_tl_faces": frame_end["traffic_light_faces_index_interval"][1]
        - frame_start["traffic_light_faces_index_interval"][0],
    }


def _append_zarr_subset(
        input_zarr: ChunkedDataset,
        output_zarr: ChunkedDataset,
        scene_index_start: int,
        scene_index_end: int,
        output_zarr_num_els: Optional[dict] = None,
) -> None:
    """
    Append a subset of input_zarr into output_zarr. To avoid appending (slow), output_zarr must be opened in write mode
    and with pre-allocated shape already. End indices of output_zarr are read from output_zarr_num_els, or 0 is assumed
    otherwise

    Args:
        input_zarr (ChunkedDataset): origin zarr in read mode
        output_zarr (ChunkedDataset): zarr already opened in write mode and with pre-allocated arrays
        scene_index_start (int): index of the first scene to copy
        scene_index_end (int): index of the last scene (excluded)
        output_zarr_num_els (Optional[dict]): if None, write starting from 0 index in the output zarr

    Returns:

    """

    # indices to assign in the destination array
    if output_zarr_num_els is None:
        idx_output_scene, idx_output_frame, idx_output_agent, idx_output_tl_face = 0, 0, 0, 0
    else:
        idx_output_scene = output_zarr_num_els["num_scenes"]
        idx_output_frame = output_zarr_num_els["num_frames"]
        idx_output_agent = output_zarr_num_els["num_agents"]
        idx_output_tl_face = output_zarr_num_els["num_tl_faces"]

    # relative indices to subtract before copying to erase input history
    idx_start_frame = input_zarr.scenes[scene_index_start]["frame_index_interval"][0]
    idx_start_agent = input_zarr.frames[idx_start_frame]["agent_index_interval"][0]
    idx_start_tl_face = input_zarr.frames[idx_start_frame]["traffic_light_faces_index_interval"][0]

    # if output_zarr_num_els is not zero we also need to add output_history
    idx_start_frame = idx_output_frame - idx_start_frame
    idx_start_agent = idx_output_agent - idx_start_agent
    idx_start_tl_face = idx_output_tl_face - idx_start_tl_face

    for idx_scene in range(scene_index_start, scene_index_end):
        # get slices from input zarr
        scenes = input_zarr.scenes[idx_scene: idx_scene + 1]
        frames = input_zarr.frames[get_frames_slice_from_scenes(*scenes)]
        agents = input_zarr.agents[get_agents_slice_from_frames(*frames[[0, -1]])]
        tl_faces = input_zarr.tl_faces[get_tl_faces_slice_from_frames(*frames[[0, -1]])]

        # fix indices
        scenes["frame_index_interval"] += idx_start_frame
        frames["agent_index_interval"] += idx_start_agent
        frames["traffic_light_faces_index_interval"] += idx_start_tl_face

        # copy from input_zarr to output_zarr
        output_zarr.scenes[idx_output_scene: idx_output_scene + len(scenes)] = scenes
        output_zarr.frames[idx_output_frame: idx_output_frame + len(frames)] = frames
        output_zarr.agents[idx_output_agent: idx_output_agent + len(agents)] = agents
        output_zarr.tl_faces[idx_output_tl_face: idx_output_tl_face + len(tl_faces)] = tl_faces

        # update output indices
        idx_output_scene += len(scenes)
        idx_output_frame += len(frames)
        idx_output_agent += len(agents)
        idx_output_tl_face += len(tl_faces)


def zarr_concat(input_zarrs: List[str], output_zarr: str) -> None:
    """
    Concat many zarr into a single one. Takes care of updating indices for frames and agents.

    Args:
        input_zarrs (List[str]): a list of paths to input zarrs
        output_zarr (str): the path to the output zarr

    Returns:

    """

    assert not os.path.exists(output_zarr), "we need to pre-allocate zarr, can't append fast"
    output_dataset = ChunkedDataset(output_zarr)

    # we need to estimate how much to allocate by reading all input zarrs lens
    # we also store them for later use
    num_els_valid_zarrs = []
    valid_zarrs = []

    tqdm_bar = tqdm(input_zarrs, desc="computing total size to allocate")
    for input_zarr in tqdm_bar:
        try:
            input_dataset = ChunkedDataset(input_zarr)
            input_dataset.open()
        except (ValueError, KeyError):
            print(f"{input_zarr} is not valid! skipping")
            continue
        num_els_valid_zarrs.append(_get_num_els_in_scene_range(input_dataset, 0, len(input_dataset.scenes)))
        valid_zarrs.append(input_zarr)

    # we can now pre-allocate the output dataset
    total_num_els: Counter = Counter()
    for num_el in num_els_valid_zarrs:
        total_num_els += Counter(num_el)
    output_dataset.initialize(**total_num_els)

    cur_num_els = Counter({"num_scenes": 0, "num_frames": 0, "num_agents": 0, "num_tl_faces": 0})
    tqdm_bar = tqdm(valid_zarrs)
    for idx, input_zarr in enumerate(tqdm_bar):
        tqdm_bar.set_description(f"working on {input_zarr}")

        input_dataset = ChunkedDataset(input_zarr)
        input_dataset.open()

        _append_zarr_subset(input_dataset, output_dataset, 0, len(input_dataset.scenes), cur_num_els)
        cur_num_els += Counter(num_els_valid_zarrs[idx])


def zarr_split(input_zarr: str, output_path: str, split_infos: List[dict]) -> List[Tuple[int, int]]:
    """
    Split the input zarr into many zarrs. Names and sizes can be passed using the split_infos arg.

    Args:
        input_zarr (str): path of the original zarr
        output_path (str): base destination path
        split_infos (List[dict]): list of dict. Each element should have `name` (final path is output_path+name)
        and `split_size_GB` with the size of the split. Last element must have `split_size_GB` set to -1 to collect
        the last part of the input_zarrr.

    Returns:
        List[Tuple[int, int]]: indices of scenes where a split occurred
    """
    input_dataset = ChunkedDataset(input_zarr)
    input_dataset.open()

    assert len(split_infos) > 0
    assert split_infos[-1]["split_size_GB"] == -1, "last split element should have split_size_GB equal to -1"

    # compute the size of the input_dataset in GB
    size_input_gb = _compute_path_size(input_zarr) / GIGABYTE
    num_scenes_input = len(input_dataset.scenes)

    # ensure the defined splits don't overspill the input dataset
    num_scenes_output = [
        int(num_scenes_input * split_info["split_size_GB"] / size_input_gb) for split_info in split_infos[:-1]
    ]
    assert sum(num_scenes_output) < num_scenes_input, "size exceed"
    num_scenes_output.append(num_scenes_input - sum(num_scenes_output))

    cur_scene = 0
    cuts_track = []  # keep track of start-end of the cuts
    tqdm_bar = tqdm(zip(num_scenes_output, split_infos))
    for num_scenes, split_info in tqdm_bar:
        start_cut = cur_scene
        end_cut = cur_scene + num_scenes
        tqdm_bar.set_description(f"cutting scenes {start_cut}-{end_cut} into {split_info['name']}")

        num_els_output = _get_num_els_in_scene_range(input_dataset, start_cut, end_cut)

        output_dataset = ChunkedDataset(str(Path(output_path) / split_info["name"]))
        output_dataset.initialize(**num_els_output)

        _append_zarr_subset(input_dataset, output_dataset, start_cut, end_cut)
        cuts_track.append((start_cut, end_cut))
        cur_scene = end_cut

    return cuts_track


def zarr_scenes_chop(input_zarr: str, output_zarr: str, num_frames_to_copy: int) -> None:
    """
    Copy `num_frames_to_keep` from each scene in input_zarr and paste them into output_zarr

    Args:
        input_zarr (str): path to the input zarr
        output_zarr (str): path to the output zarr
        num_frames_to_copy (int): how many frames to copy from the start of each scene

    Returns:

    """
    input_dataset = ChunkedDataset(input_zarr)
    input_dataset.open()

    # check we can actually copy the frames we want from each scene
    assert np.all(np.diff(input_dataset.scenes["frame_index_interval"], 1) > num_frames_to_copy), "not enough frames"

    output_dataset = ChunkedDataset(output_zarr)
    output_dataset.initialize()

    # current indices where to copy in the output_dataset
    cur_scene_idx, cur_frame_idx, cur_agent_idx, cur_tl_face_idx = 0, 0, 0, 0

    for idx in tqdm(range(len(input_dataset.scenes)), desc="copying"):
        # get data and immediately chop frames, agents and traffic lights
        scene = input_dataset.scenes[idx]
        first_frame_idx = scene["frame_index_interval"][0]

        frames = input_dataset.frames[first_frame_idx: first_frame_idx + num_frames_to_copy]
        agents = input_dataset.agents[get_agents_slice_from_frames(*frames[[0, -1]])]
        tl_faces = input_dataset.tl_faces[get_tl_faces_slice_from_frames(*frames[[0, -1]])]

        # reset interval relative to our output (subtract current history and add output history)
        scene["frame_index_interval"][0] = cur_frame_idx
        scene["frame_index_interval"][1] = cur_frame_idx + num_frames_to_copy  # address for less frames

        frames["agent_index_interval"] += cur_agent_idx - frames[0]["agent_index_interval"][0]
        frames["traffic_light_faces_index_interval"] += (
            cur_tl_face_idx - frames[0]["traffic_light_faces_index_interval"][0]
        )

        # write in dest using append (slow)
        output_dataset.scenes.append(scene[None, ...])  # need 2D array to concatenate
        output_dataset.frames.append(frames)
        output_dataset.agents.append(agents)
        output_dataset.tl_faces.append(tl_faces)

        # increase indices in output
        cur_scene_idx += len(scene)
        cur_frame_idx += len(frames)
        cur_agent_idx += len(agents)
        cur_tl_face_idx += len(tl_faces)



import bisect
import warnings
from functools import partial
from typing import Optional

import numpy as np
from torch.utils.data import Dataset

from ..data import ChunkedDataset, get_frames_slice_from_scenes
from ..kinematic import Perturbation
from ..rasterization import Rasterizer, RenderContext
from ..sampling import generate_agent_sample
from .utils import convert_str_to_fixed_length_tensor


class EgoDataset(Dataset):
    def __init__(
            self,
            cfg: dict,
            zarr_dataset: ChunkedDataset,
            rasterizer: Rasterizer,
            perturbation: Optional[Perturbation] = None,
    ):
        """
        Get a PyTorch dataset object that can be used to train DNN

        Args:
            cfg (dict): configuration file
            zarr_dataset (ChunkedDataset): the raw zarr dataset
            rasterizer (Rasterizer): an object that support rasterisation around an agent (AV or not)
            perturbation (Optional[Perturbation]): an object that takes care of applying trajectory perturbations.
None if not desired
        """
        self.perturbation = perturbation
        self.cfg = cfg
        self.dataset = zarr_dataset
        self.rasterizer = rasterizer

        self.cumulative_sizes = self.dataset.scenes["frame_index_interval"][:, 1]

        render_context = RenderContext(
            raster_size_px=np.array(cfg["raster_params"]["raster_size"]),
            pixel_size_m=np.array(cfg["raster_params"]["pixel_size"]),
            center_in_raster_ratio=np.array(cfg["raster_params"]["ego_center"]),
            set_origin_to_bottom=cfg["raster_params"]["set_origin_to_bottom"],
        )

        # build a partial so we don't have to access cfg each time
        self.sample_function = partial(
            generate_agent_sample,
            render_context=render_context,
            history_num_frames=cfg["model_params"]["history_num_frames"],
            future_num_frames=cfg["model_params"]["future_num_frames"],
            step_time=cfg["model_params"]["step_time"],
            filter_agents_threshold=cfg["raster_params"]["filter_agents_threshold"],
            rasterizer=rasterizer,
            perturbation=perturbation,
        )

    def __len__(self) -> int:
        """
        Get the number of available AV frames

        Returns:
            int: the number of elements in the dataset
        """
        return len(self.dataset.frames)

    def get_frame(self, scene_index: int, state_index: int, track_id: Optional[int] = None) -> dict:
        """
        A utility function to get the rasterisation and trajectory target for a given agent in a given frame

        Args:
            scene_index (int): the index of the scene in the zarr
            state_index (int): a relative frame index in the scene
            track_id (Optional[int]): the agent to rasterize or None for the AV
        Returns:
            dict: the rasterised image in (Cx0x1) if the rast is not None, the target trajectory
            (position and yaw) along with their availability, the 2D matrix to center that agent,
            the agent track (-1 if ego) and the timestamp

        """
        frames = self.dataset.frames[get_frames_slice_from_scenes(self.dataset.scenes[scene_index])]

        tl_faces = self.dataset.tl_faces
        try:
            if self.cfg["raster_params"]["disable_traffic_light_faces"]:
                tl_faces = np.empty(0, dtype=self.dataset.tl_faces.dtype)  # completely disable traffic light faces
        except KeyError:
            warnings.warn(
                "disable_traffic_light_faces not found in config, this will raise an error in the future",
                RuntimeWarning,
                stacklevel=2,
            )
        data = self.sample_function(state_index, frames, self.dataset.agents, tl_faces, track_id)

        # add information only, so that all data keys are always preserved
        data["scene_index"] = scene_index
        data["host_id"] = np.uint8(convert_str_to_fixed_length_tensor(self.dataset.scenes[scene_index]["host"]).cpu())
        data["timestamp"] = frames[state_index]["timestamp"]
        data["track_id"] = np.int64(-1 if track_id is None else track_id)  # always a number to avoid crashing torch
        data["world_to_image"] = data["raster_from_world"]  # TODO deprecate

        # when rast is None, image could be None. In that case we remove the key
        if data["image"] is not None:
            data["image"] = data["image"].transpose(2, 0, 1)  # 0,1,C -> C,0,1
        else:
            del data["image"]

        return data

    def __getitem__(self, index: int) -> dict:
        """
        Function called by Torch to get an element

        Args:
            index (int): index of the element to retrieve

        Returns: please look get_frame signature and docstring

        """
        if index < 0:
            if -index > len(self):
                raise ValueError("absolute value of index should not exceed dataset length")
            index = len(self) + index

        scene_index = bisect.bisect_right(self.cumulative_sizes, index)

        if scene_index == 0:
            state_index = index
        else:
            state_index = index - self.cumulative_sizes[scene_index - 1]
        return self.get_frame(scene_index, state_index)

    def get_scene_dataset(self, scene_index: int) -> "EgoDataset":
        """
        Returns another EgoDataset dataset where the underlying data can be modified.
        This is possible because, even if it supports the same interface, this dataset is np.ndarray based.

        Args:
            scene_index (int): the scene index of the new dataset

        Returns:
            EgoDataset: A valid EgoDataset dataset with a copy of the data

        """
        dataset = self.dataset.get_scene_dataset(scene_index)
        return EgoDataset(self.cfg, dataset, self.rasterizer, self.perturbation)

    def get_scene_indices(self, scene_idx: int) -> np.ndarray:
        """
        Get indices for the given scene. EgoDataset iterates over frames, so this is just a matter
        of finding the scene boundaries.
        Args:
            scene_idx (int): index of the scene

        Returns:
            np.ndarray: indices that can be used for indexing with __getitem__
        """
        scenes = self.dataset.scenes
        assert scene_idx < len(scenes), f"scene_idx {scene_idx} is over len {len(scenes)}"
        return np.arange(*scenes[scene_idx]["frame_index_interval"])

    def get_frame_indices(self, frame_idx: int) -> np.ndarray:
        """
        Get indices for the given frame. EgoDataset iterates over frames, so this will be a single element
        Args:
            frame_idx (int): index of the scene

        Returns:
            np.ndarray: indices that can be used for indexing with __getitem__
        """
        frames = self.dataset.frames
        assert frame_idx < len(frames), f"frame_idx {frame_idx} is over len {len(frames)}"
        return np.asarray((frame_idx,), dtype=np.int64)

    def __str__(self) -> str:
        return self.dataset.__str__()



from abc import ABC, abstractmethod
from typing import Any, Dict, List

from l5kit.cle import composite_metrics, metrics, validators
from l5kit.cle.closed_loop_evaluator import ClosedLoopEvaluator, EvaluationPlan
from l5kit.simulation.unroll import SimulationOutputCLE


class BaseMetricSet(ABC):
    """Base class interface for the metric sets."""
    #: The prefix that will identify this metric
    metric_prefix: str
    evaluator: ClosedLoopEvaluator

    @abstractmethod
    def evaluate(self, sim_outputs: List[SimulationOutputCLE]) -> None:
        """Run the CLE (Closed-loop Evaluator) on simulated scenes.

        :param sim_outputs: outputs from the simulator.
        """
        raise NotImplementedError

    @abstractmethod
    def get_results(self) -> Dict[str, Any]:
        """Perform all required aggregations and returns a dictionary
        composed by all results."""
        raise NotImplementedError


class L5MetricSet(BaseMetricSet):
    """This class is responsible for computing a set of metric parametrization for the L5Kit.

    :param metric_prefix: this is a prefix that will identify the metric set being used.
    """

    def __init__(self, metric_prefix: str = "L5") -> None:
        self.metric_prefix = metric_prefix
        metric_list = self.build_metrics()
        validators_list = self.build_validators()
        composite_metric_list = self.build_composite_metrics()
        validators_intervention = self.get_validator_interventions()
        # Builds the evaluation plan and run the closed loop evaluator
        self.evaluation_plan = EvaluationPlan(metric_list, validators_list,
                                              composite_metric_list, validators_intervention)
        self.evaluator = ClosedLoopEvaluator(self.evaluation_plan)

    @abstractmethod
    def build_metrics(self) -> List[metrics.SupportsMetricCompute]:
        """Returns a list of metrics that will be computed.
        """
        raise NotImplementedError

    def build_validators(self) -> List[validators.SupportsMetricValidate]:
        """Returns a list of validators that will operate on the computed metrics.
        """
        return []

    def get_validator_interventions(self) -> List[str]:
        """Returns a list of validators that are considered an intervention.
        """
        return []

    def build_composite_metrics(self) -> List[composite_metrics.SupportsCompositeMetricCompute]:
        """Return a list of composite metrics that should be computed. Composite
        metrics are metrics that depend upon metrics and validator results.
        """
        return []

    def evaluate(self, sim_outputs: List[SimulationOutputCLE]) -> None:
        """Run the CLE (Closed-loop Evaluator) on simulated scenes.

        :param sim_outputs: outputs from the simulator.
        """
        self.evaluator.evaluate(sim_outputs)

    def reset(self) -> None:
        """Reset the current state of the CLE (Closed-loop Evaluator).
        """
        self.evaluator.reset()

    def get_results(self) -> Dict[str, Any]:
        """Perform all required aggregations and returns a dictionary composed by all results.
        """
        raise NotImplementedError



from typing import Optional, Tuple

import cv2
import numpy as np

from l5kit.geometry import transform_point, transform_points


PREDICTED_POINTS_COLOR = (0, 255, 255)
TARGET_POINTS_COLOR = (255, 0, 255)
REFERENCE_TRAJ_COLOR = (255, 255, 0)
#  Arrows represent position + orientation.
ARROW_LENGTH_IN_PIXELS = 2
ARROW_THICKNESS_IN_PIXELS = 1
ARROW_TIP_LENGTH_IN_PIXELS = 1.8


def draw_arrowed_line(on_image: np.ndarray, position: np.ndarray, yaw: float, rgb_color: Tuple[int, int, int]) -> None:
    """
    Draw a single arrowed line in an RGB image
    Args:
        on_image (np.ndarray): the RGB image to draw onto
        position (np.ndarray): the starting position of the arrow
        yaw (float): the arrow orientation
        rgb_color (Tuple[int, int, int]): the arrow color

    Returns: None

    """
    start_pixel = np.array(position[:2])

    rot = np.eye(3)
    rot[:-1] = cv2.getRotationMatrix2D((0, 0), np.degrees(-yaw), 1.0)  # minus here because of cv2 rotations convention
    end_pixel = start_pixel + transform_point(np.asarray([ARROW_LENGTH_IN_PIXELS, 0]), rot)

    cv2.arrowedLine(
        on_image,
        tuple(start_pixel.astype(np.int32)),
        tuple(end_pixel.astype(np.int32)),
        rgb_color,
        thickness=ARROW_THICKNESS_IN_PIXELS,
        tipLength=ARROW_TIP_LENGTH_IN_PIXELS,
    )


def draw_trajectory(
        on_image: np.ndarray,
        positions: np.ndarray,
        rgb_color: Tuple[int, int, int],
        radius: int = 1,
        yaws: Optional[np.ndarray] = None,
) -> None:
    """
    Draw a trajectory on oriented arrow onto an RGB image
    Args:
        on_image (np.ndarray): the RGB image to draw onto
        positions (np.ndarray): pixel coordinates in the image space (not displacements) (Nx2)
        rgb_color (Tuple[int, int, int]): the trajectory RGB color
        radius (int): radius of the circle
        yaws (Optional[np.ndarray]): yaws in radians (N) or None to disable yaw visualisation

    Returns: None

    """
    if yaws is not None:
        assert len(yaws) == len(positions)
        for pos, yaw in zip(positions, yaws):
            pred_waypoint = pos[:2]
            pred_yaw = float(yaw[0])
            draw_arrowed_line(on_image, pred_waypoint, pred_yaw, rgb_color)
    else:
        for pos in positions:
            pred_waypoint = pos[:2]
            cv2.circle(on_image, tuple(pred_waypoint.astype(np.int)), radius, rgb_color, -1)


def draw_reference_trajectory(on_image: np.ndarray, world_to_pixel: np.ndarray, positions: np.ndarray) -> None:
    """
    Draw a trajectory (as points) onto the image
    Args:
        on_image (np.ndarray): the RGB image to draw onto
        world_to_pixel (np.ndarray): 3x3 matrix from meters to ego pixel space
        positions (np.ndarray): positions as 2D absolute meters coordinates

    Returns: None

    """
    positions_in_pixel_space = transform_points(np.array(positions), world_to_pixel)
    #  we clip the positions to be within the image
    mask = np.all(positions_in_pixel_space > (0.0, 0.0), 1) * np.all(positions_in_pixel_space < on_image.shape[:2], 1)
    positions_in_pixel_space = positions_in_pixel_space[mask]
    for pos in positions_in_pixel_space:
        cv2.circle(on_image, tuple(np.floor(pos).astype(np.int32)), 1, REFERENCE_TRAJ_COLOR, -1)



from abc import abstractmethod
from collections import defaultdict
from enum import IntEnum
from typing import Any, DefaultDict, Dict, List, Type

import torch
from typing_extensions import Protocol

from l5kit.cle import metrics, validators
from l5kit.cle.metrics import SupportsMetricCompute
from l5kit.simulation.unroll import SimulationOutputCLE


class SupportsCompositeMetricCompute(Protocol):
    """Protocol supporting the computation method for metrics."""

    #: Name of the composite metric
    composite_metric_name: str
    #: List of names for metrics this composite metric depends on
    requires_metric: List[str]
    #: List of validators that this composite metric depends on
    requires_validator: List[str]

    @abstractmethod
    def compute(self, metric_results: Dict[str, torch.Tensor],
                validation_results: Dict[str, validators.ValidatorOutput],
                simulation_output: SimulationOutputCLE) -> float:
        """Method that supports the computation of the composite metric. This
        metric should return a single float per scene.

        :param metric_results: results of the metrics required
        :tensor metric_results: [N], where N is the number of frames in the scene
        :param validation_results: results from the validators required
        :param simulation_output: output from the closed-loop simulator
        :returns: a float result
        """
        raise NotImplementedError


class PassedDrivenMilesCompositeMetric(SupportsCompositeMetricCompute):
    """This composite metric will compute the "passed driven miles", which
    represents the sum of driven miles up to a first intervention frame
    that is detected based on the list of specified validators.

    :param intervention_validators: an intervention is a defined by any
                                    failed frame from any of the specified
                                    validators.
    :param driven_miles_metric: the metric that should be used as driven
                                miles (defaults to simulated driven miles),
                                but can also use the driven miles by the
                                log replay.
    :param ignore_entire_scene: if the entire driven miles from the scene
                                should be ignored when there was an intervention.
    """
    #: Name of the composite metric
    composite_metric_name: str
    #: List of names for metrics this composite metric depends on
    requires_metric: List[str]
    #: List of validators that this composite metric depends on
    requires_validator: List[str]

    def __init__(self, composite_metric_name: str,
                 intervention_validators: List[str],
                 driven_miles_metric: Type[SupportsMetricCompute] =
                 metrics.SimulatedDrivenMilesMetric,
                 ignore_entire_scene: bool = False):
        self.composite_metric_name = composite_metric_name
        self.requires_metric = [driven_miles_metric.metric_name]
        self.requires_validator = list(set(intervention_validators))
        self.ignore_entire_scene = ignore_entire_scene

    def compute(self, metric_results: Dict[str, torch.Tensor],
                validation_results: Dict[str, validators.ValidatorOutput],
                simulation_output: SimulationOutputCLE) -> float:
        """Computes the driven miles until the first intervention
        is found (first failed frame).

        :param metric_results: results of the metrics required
        :tensor metric_results: [N], where N is the number of frames in the scene
        :param validation_results: results from the validators required
        :param simulation_output: output from the closed-loop simulator
        :returns: passed driven miles
        """
        driven_miles_result = metric_results[self.requires_metric[0]]

        min_all_frame_failed: List[int] = [driven_miles_result.size(0)]
        for validator_name in self.requires_validator:
            validator_results = validation_results[validator_name]
            if len(validator_results.failed_frames) > 0:
                # If ignore the entire scene is enabled, we just
                # return 0 passed miles
                if self.ignore_entire_scene:
                    return 0.0

                # ... otherwise, we check for the minimum failed
                # frame to compute the proportional driven miles
                min_frame_failed = min(validator_results.failed_frames)
                min_all_frame_failed.append(min_frame_failed)

        min_frame_failed = min(min_all_frame_failed)
        passed_driven_miles = driven_miles_result[:min_frame_failed].sum()
        passed_driven_miles_cpu = passed_driven_miles.cpu().item()
        return float(passed_driven_miles_cpu)


class DrivenMilesCompositeMetric(SupportsCompositeMetricCompute):
    """Composite metric to accumulate the total driven miles.

    :param composite_metric_name: name of the composite metric
    :param driven_miles_metric: the metric that should be used
                                to accumulate its values, defaults
                                to the simulated driven miles
    """
    #: Name of the composite metric
    composite_metric_name: str
    #: List of names for metrics this composite metric depends on
    requires_metric: List[str]
    #: List of validators that this composite metric depends on
    requires_validator: List[str]

    def __init__(self, composite_metric_name: str,
                 driven_miles_metric: Type[SupportsMetricCompute] =
                 metrics.SimulatedDrivenMilesMetric):
        self.composite_metric_name = composite_metric_name
        self.driven_miles_metric_name = driven_miles_metric.metric_name
        self.requires_metric = [self.driven_miles_metric_name]
        self.requires_validator = []

    def compute(self, metric_results: Dict[str, torch.Tensor],
                validation_results: Dict[str, validators.ValidatorOutput],
                simulation_output: SimulationOutputCLE) -> float:
        driven_miles = metric_results[self.driven_miles_metric_name].sum()
        driven_miles_cpu = driven_miles.cpu().item()
        return float(driven_miles_cpu)


class SupportsCompositeMetricAggregation(Protocol):
    """Protocol supporting the composite metric aggregation. This aggregator
    is responsible for aggregating results from the the composite metrics and
    also doing a reduction across multiple distributed nodes.
    """

    def aggregate(self, scene_composite_metric_results:
                  Dict[int, Dict[str, float]]) -> Dict[str, Any]:
        """This method will aggregate scenes locally and then will
        do the reduction step to aggregate data across distributed
        nodes.

        :param scene_composite_metric_results: results from composite metric
                                               outputs per scene
        :returns: any result indexed by the composite metric name.
        """
        raise NotImplementedError


class ReduceMode(IntEnum):
    """Reduction modes supported by the composite metric aggregator."""
    #: Summation of results locally and across nodes
    SUM = 0


class CompositeMetricAggregator(SupportsCompositeMetricAggregation):
    """Aggregates composite metric results locally and across multiple
    distributed nodes.

    :param reduce_mode: the reduce operation to be applied both into
                        local scenes and also across distributed
                        nodes.
    """
    def __init__(self, reduce_mode: ReduceMode = ReduceMode.SUM):
        # Supporting only ReduceMode.SUM for the moment
        if reduce_mode not in set(ReduceMode):
            raise ValueError(f"Reduce mode {reduce_mode} not implemented.")
        self.reduce_mode = reduce_mode

    def aggregate_scenes(self, scene_composite_metric_results:
                         Dict[int, Dict[str, float]]) -> Dict[str, Any]:
        """Aggregate the scenes locally on each node. This method will just
        sum the number of invalid scenes across all scenes in the node.

        :param scene_composite_metric_results: results from composite metric
                                               outputs per scene
        :returns: a dictionary with the composite metric name as keys
                  and the reduction of the composite metrics
        """
        if self.reduce_mode == ReduceMode.SUM:
            aggregation: DefaultDict[str, float] = defaultdict(float)
            for _, cm_dict in scene_composite_metric_results.items():
                for validator_name, cm_output in cm_dict.items():
                    aggregation[validator_name] += cm_output
            aggregation_torch = {k: torch.as_tensor(v) for k, v in aggregation.items()}
            return aggregation_torch
        return {}

    def aggregate(self, scene_composite_metric_results:
                  Dict[int, Dict[str, float]]) -> Dict[str, Any]:
        """This method will aggregate scenes locally and then will
        do the reduction step to aggregate data across distributed
        nodes.

        :param scene_composite_metric_results: results from composite metric
                                               outputs per scene
        :returns: result indexed by the composite metric name.
        """
        agg_scenes = self.aggregate_scenes(scene_composite_metric_results)
        return agg_scenes



from collections import defaultdict
from dataclasses import dataclass
from typing import Any, DefaultDict, Dict, List, NamedTuple, Optional

import gym
import numpy as np
import torch
from gym import spaces
from gym.utils import seeding

from l5kit.configs import load_config_data
from l5kit.data import ChunkedDataset, LocalDataManager
from l5kit.dataset import EgoDataset
from l5kit.environment.kinematic_model import KinematicModel, UnicycleModel
from l5kit.environment.reward import L2DisplacementYawReward, Reward
from l5kit.environment.utils import (calculate_kinematic_rescale_params, calculate_non_kinematic_rescale_params,
                                     KinematicActionRescaleParams, NonKinematicActionRescaleParams)
from l5kit.rasterization import build_rasterizer
from l5kit.simulation.dataset import SimulationConfig, SimulationDataset
from l5kit.simulation.unroll import (ClosedLoopSimulator, ClosedLoopSimulatorModes, SimulationOutputCLE,
                                     UnrollInputOutput)


@dataclass
class SimulationConfigGym(SimulationConfig):
    """Defines the default parameters used for the simulation of ego and agents around it in L5Kit Gym.
    Note: num_simulation_steps should be eps_length + 1
    This is because we (may) require to extract the initial speed of the vehicle for the kinematic model
    The speed at start_frame_idx is always 0 (not indicative of the true current speed).
    We therefore simulate the episode from (start_frame_idx + 1, start_frame_idx + eps_length + 1)

    :param use_ego_gt: whether to use GT annotations for ego instead of model's outputs
    :param use_agents_gt: whether to use GT annotations for agents instead of model's outputs
    :param disable_new_agents: whether to disable agents that are not returned at start_frame_index
    :param distance_th_far: if a tracked agent is closed than this value to ego, it will be controlled
    :param distance_th_close: if a new agent is closer than this value to ego, it will be controlled
    :param start_frame_index: the start index of the simulation
    :param num_simulation_steps: the number of step to simulate
    """
    use_ego_gt: bool = False
    use_agents_gt: bool = True
    disable_new_agents: bool = False
    distance_th_far: float = 30.0
    distance_th_close: float = 15.0
    start_frame_idx: int = 0
    num_simulation_steps: int = 33


class EpisodeOutputGym(SimulationOutputCLE):
    """This object holds information regarding the simulation output at the end of an episode
    in the gym-compatible L5Kit environment. The output can be used to
    calculate quantitative metrics and provide qualitative visualization.

    :param scene_id: the scene indices
    :param sim_dataset: the simulation dataset
    :param ego_ins_outs: all inputs and outputs for ego (each frame of each scene has only one)
    :param agents_ins_outs: all inputs and outputs for agents (multiple per frame in a scene)
    """

    def __init__(self, scene_id: int, sim_dataset: SimulationDataset,
                 ego_ins_outs: DefaultDict[int, List[UnrollInputOutput]],
                 agents_ins_outs: DefaultDict[int, List[List[UnrollInputOutput]]]):
        """Constructor method
        """
        super(EpisodeOutputGym, self).__init__(scene_id, sim_dataset, ego_ins_outs, agents_ins_outs)

        # Required for Bokeh Visualizer
        simulated_dataset = sim_dataset.scene_dataset_batch[scene_id]
        self.tls_frames = simulated_dataset.dataset.tl_faces
        self.agents_th = simulated_dataset.cfg["raster_params"]["filter_agents_threshold"]


class GymStepOutput(NamedTuple):
    """The output dict of gym env.step

    :param obs: the next observation on performing environment step
    :param reward: the reward of the current step
    :param done: flag to indicate end of episode
    :param info: additional information
    """
    obs: Dict[str, np.ndarray]
    reward: float
    done: bool
    info: Dict[str, Any]


class L5Env(gym.Env):
    """Custom Environment of L5 Kit that can be registered in OpenAI Gym.

    :param env_config_path: path to the L5Kit environment configuration file
    :param dmg: local data manager object
    :param simulation_cfg: configuration of the L5Kit closed loop simulator
    :param reward: calculates the reward for the gym environment
    :param cle: flag to enable close loop environment updates
    :param rescale_action: flag to rescale the model action back to the un-normalized action space
    :param use_kinematic: flag to use the kinematic model
    :param kin_model: the kinematic model
    :param return_info: flag to return info when a episode ends
    """

    def __init__(self, env_config_path: Optional[str] = None, dmg: Optional[LocalDataManager] = None,
                 sim_cfg: Optional[SimulationConfig] = None,
                 reward: Optional[Reward] = None, cle: bool = True, rescale_action: bool = True,
                 use_kinematic: bool = False, kin_model: Optional[KinematicModel] = None,
                 reset_scene_id: Optional[int] = None, return_info: bool = False) -> None:
        """Constructor method
        """
        super(L5Env, self).__init__()

        # Required to register environment
        if env_config_path is None:
            return

        # env config
        dm = dmg if dmg is not None else LocalDataManager(None)
        cfg = load_config_data(env_config_path)
        self.step_time = cfg["model_params"]["step_time"]

        # rasterisation
        rasterizer = build_rasterizer(cfg, dm)
        raster_size = cfg["raster_params"]["raster_size"][0]
        n_channels = rasterizer.num_channels()

        # init dataset
        train_zarr = ChunkedDataset(dm.require(cfg["train_data_loader"]["key"])).open()
        self.dataset = EgoDataset(cfg, train_zarr, rasterizer)

        # Define action and observation space
        # Continuous Action Space: gym.spaces.Box (X, Y, Yaw * number of future states)
        self.action_space = spaces.Box(low=-1, high=1, shape=(3, ))

        # Observation Space: gym.spaces.Dict (image: [n_channels, raster_size, raster_size])
        obs_shape = (n_channels, raster_size, raster_size)
        self.observation_space = spaces.Dict({'image': spaces.Box(low=0, high=1, shape=obs_shape, dtype=np.float32)})

        # Simulator Config within Gym
        self.sim_cfg = sim_cfg if sim_cfg is not None else SimulationConfigGym()
        self.simulator = ClosedLoopSimulator(self.sim_cfg, self.dataset, device=torch.device("cpu"),
                                             mode=ClosedLoopSimulatorModes.GYM)

        self.reward = reward if reward is not None else L2DisplacementYawReward()

        self.max_scene_id = cfg["gym_params"]["max_scene_id"]
        self.overfit = cfg["gym_params"]["overfit"]
        if self.overfit:
            self.overfit_scene_id = cfg["gym_params"]["overfit_id"]

        self.cle = cle
        self.rescale_action = rescale_action
        self.use_kinematic = use_kinematic

        if self.use_kinematic:
            self.kin_model = kin_model if kin_model is not None else UnicycleModel()
            self.kin_rescale = self._get_kin_rescale_params()
        else:
            self.non_kin_rescale = self._get_non_kin_rescale_params()

        # If not None, reset_scene_id is the scene_id that will be rolled out when reset is called
        self.reset_scene_id = reset_scene_id
        if self.overfit:
            self.reset_scene_id = self.overfit_scene_id

        # flag to decide whether to return any info at end of episode
        # helps to limit the IPC
        self.return_info = return_info

        self.seed()

    def seed(self, seed: int = None) -> List[int]:
        """ Generate the random seed.

        :param seed: the seed integer
        :return: the output random seed
        """
        self.np_random, seed = seeding.np_random(seed)
        # TODO : add a torch seed for future
        return [seed]

    def reset(self) -> Dict[str, np.ndarray]:
        """ Resets the environment and outputs first frame of a new scene sample.

        :return: the observation of first frame of sampled scene index
        """
        # Define in / outs for new episode scene
        self.agents_ins_outs: DefaultDict[int, List[List[UnrollInputOutput]]] = defaultdict(list)
        self.ego_ins_outs: DefaultDict[int, List[UnrollInputOutput]] = defaultdict(list)

        if self.reset_scene_id is not None:
            self.scene_index = self.reset_scene_id
        else:  # Sample a scene
            self.scene_index = self.np_random.randint(0, self.max_scene_id)
        self.sim_dataset = SimulationDataset.from_dataset_indices(self.dataset, [self.scene_index], self.sim_cfg)

        # Reset CLE evaluator
        self.reward.reset()

        # Output first observation
        self.frame_index = 1  # Frame_index 1 has access to the true ego speed
        ego_input = self.sim_dataset.rasterise_frame_batch(self.frame_index)
        self.ego_input_dict = {k: np.expand_dims(v, axis=0) for k, v in ego_input[0].items()}

        # Reset Kinematic model
        if self.use_kinematic:
            init_kin_state = np.array([0.0, 0.0, 0.0, self.step_time * ego_input[0]['curr_speed']])
            self.kin_model.reset(init_kin_state)

        # Only output the image attribute
        obs = {'image': ego_input[0]["image"]}
        return obs

    def step(self, action: np.ndarray) -> GymStepOutput:
        """Inputs the action, updates the environment state and outputs the next frame.

        :param action: the action to perform on current state/frame
        :return: the namedTuple comprising the (next observation, reward, done, info)
            based on the current action
        """
        frame_index = self.frame_index
        next_frame_index = frame_index + 1
        episode_over = next_frame_index == (len(self.sim_dataset) - 1)

        # EGO
        if not self.sim_cfg.use_ego_gt:
            action = self._rescale_action(action)
            ego_output = self._convert_action_to_ego_output(action)
            self.ego_output_dict = ego_output

            if self.cle:
                # In closed loop training, the raster is updated according to predicted ego positions.
                self.simulator.update_ego(self.sim_dataset, next_frame_index, self.ego_input_dict, self.ego_output_dict)

            ego_frame_in_out = self.simulator.get_ego_in_out(self.ego_input_dict, self.ego_output_dict,
                                                             self.simulator.keys_to_exclude)
            self.ego_ins_outs[self.scene_index].append(ego_frame_in_out[self.scene_index])

        # generate simulated_outputs
        simulated_outputs = SimulationOutputCLE(self.scene_index, self.sim_dataset, self.ego_ins_outs,
                                                self.agents_ins_outs)

        # reward calculation
        reward = self.reward.get_reward(self.frame_index, [simulated_outputs])

        # done is True when episode ends
        done = episode_over

        # Optionally we can pass additional info
        # We are using "info" to output rewards and simulated outputs (during evaluation)
        info: Dict[str, Any]
        info = {'reward_tot': reward["total"], 'reward_dist': reward["distance"], 'reward_yaw': reward["yaw"]}
        if done and self.return_info:
            info = {"sim_outs": self.get_episode_outputs(), "reward_tot": reward["total"],
                    "reward_dist": reward["distance"], "reward_yaw": reward["yaw"]}

        # Get next obs
        self.frame_index += 1
        obs = self._get_obs(self.frame_index, episode_over)

        # return obs, reward, done, info
        return GymStepOutput(obs, reward["total"], done, info)

    def get_episode_outputs(self) -> List[EpisodeOutputGym]:
        """Generate and return the outputs at the end of the episode.

        :return: List of episode outputs
        """
        episode_outputs = [EpisodeOutputGym(self.scene_index, self.sim_dataset, self.ego_ins_outs,
                                            self.agents_ins_outs)]
        return episode_outputs

    def render(self) -> None:
        """Render a frame during the simulation
        """
        raise NotImplementedError

    def _get_obs(self, frame_index: int, episode_over: bool) -> Dict[str, np.ndarray]:
        """Get the observation corresponding to a given frame index in the scene.

        :param frame_index: the index of the frame which provides the observation
        :param episode_over: flag to determine if the episode is over
        :return: the observation corresponding to the frame index
        """
        if episode_over:
            frame_index = 0  # Dummy final obs (when episode_over)

        ego_input = self.sim_dataset.rasterise_frame_batch(frame_index)
        self.ego_input_dict = {k: np.expand_dims(v, axis=0) for k, v in ego_input[0].items()}
        obs = {"image": ego_input[0]["image"]}
        return obs

    def _rescale_action(self, action: np.ndarray) -> np.ndarray:
        """Rescale the input action back to the un-normalized action space. PPO and related algorithms work well
        with normalized action spaces. The environment receives a normalized action and we un-normalize it back to
        the original action space for environment updates.

        :param action: the normalized action
        :return: the unnormalized action
        """
        if self.rescale_action:
            if self.use_kinematic:
                action[0] = self.kin_rescale.steer_scale * action[0]
                action[1] = self.kin_rescale.acc_scale * action[1]
            else:
                action[0] = self.non_kin_rescale.x_mu + self.non_kin_rescale.x_scale * action[0]
                action[1] = self.non_kin_rescale.y_mu + self.non_kin_rescale.y_scale * action[1]
                action[2] = self.non_kin_rescale.yaw_mu + self.non_kin_rescale.yaw_scale * action[2]
        return action

    def _get_kin_rescale_params(self) -> KinematicActionRescaleParams:
        """Determine the action un-normalization parameters for the kinematic model
        from the current dataset in the L5Kit environment.

        :return: Tuple of the action un-normalization parameters for kinematic model
        """
        scene_ids = list(range(self.max_scene_id + 1)) if not self.overfit else [self.overfit_scene_id]
        sim_dataset = SimulationDataset.from_dataset_indices(self.dataset, scene_ids, self.sim_cfg)
        return calculate_kinematic_rescale_params(sim_dataset)

    def _get_non_kin_rescale_params(self) -> NonKinematicActionRescaleParams:
        """Determine the action un-normalization parameters for the non-kinematic model
        from the current dataset in the L5Kit environment.

        :return: Tuple of the action un-normalization parameters for non-kinematic model
        """
        scene_ids = list(range(self.max_scene_id + 1)) if not self.overfit else [self.overfit_scene_id]
        sim_dataset = SimulationDataset.from_dataset_indices(self.dataset, scene_ids, self.sim_cfg)
        return calculate_non_kinematic_rescale_params(sim_dataset)

    def _convert_action_to_ego_output(self, action: np.ndarray) -> Dict[str, np.ndarray]:
        """Convert the input action into ego output format.

        :param action: the input action provided by policy
        :return: action in ego output format, a numpy dict with keys 'positions' and 'yaws'
        """
        if self.use_kinematic:
            data_dict = self.kin_model.update(action[:2])
        else:
            # [batch_size=1, num_steps, (X, Y, yaw)]
            data = action.reshape(1, 1, 3)
            pred_positions = data[:, :, :2]
            # [batch_size, num_steps, 1->(yaw)]
            pred_yaws = data[:, :, 2:3]
            data_dict = {"positions": pred_positions, "yaws": pred_yaws}
        return data_dict



import numpy as np

from .zarr_dataset import SCENE_DTYPE


def get_combined_scenes(scenes: np.ndarray) -> np.ndarray:
    """Takes as input an np.ndarray or zarr array with scenes, and combines scenes that follow up
eachother perfectly (i.e. from a single recording by the same host). Returns an np.ndarray
of combined scenes.

    Arguments:
        scenes (np.ndarray): scenes

    Returns:
        np.ndarray -- combined scenes where followup scenes have been merged.
    """

    if len(scenes) == 0:
        return scenes

    combined_scenes = []
    start_scene = scenes[0]
    end_scene = scenes[0]

    for s in scenes[1:]:
        time_diff = s["start_time"] - end_scene["end_time"]

        if time_diff != 0 or s["host"] != start_scene["host"]:
            interval = np.array([start_scene["frame_index_interval"][0], end_scene["frame_index_interval"][1]])

            combined_scene = (interval, start_scene["host"], start_scene["start_time"], end_scene["end_time"])
            combined_scenes.append(combined_scene)
            start_scene = s

        end_scene = s

    # append the last remaining scene
    combined_scenes.append(
        (
            np.array([start_scene["frame_index_interval"][0], end_scene["frame_index_interval"][1]]),
            start_scene["host"],
            start_scene["start_time"],
            end_scene["end_time"],
        )
    )

    combined_scenes = np.array(combined_scenes, dtype=SCENE_DTYPE)
    return combined_scenes



PERCEPTION_LABELS = [
    "PERCEPTION_LABEL_NOT_SET",
    "PERCEPTION_LABEL_UNKNOWN",
    "PERCEPTION_LABEL_DONTCARE",
    "PERCEPTION_LABEL_CAR",
    "PERCEPTION_LABEL_VAN",
    "PERCEPTION_LABEL_TRAM",
    "PERCEPTION_LABEL_BUS",
    "PERCEPTION_LABEL_TRUCK",
    "PERCEPTION_LABEL_EMERGENCY_VEHICLE",
    "PERCEPTION_LABEL_OTHER_VEHICLE",
    "PERCEPTION_LABEL_BICYCLE",
    "PERCEPTION_LABEL_MOTORCYCLE",
    "PERCEPTION_LABEL_CYCLIST",
    "PERCEPTION_LABEL_MOTORCYCLIST",
    "PERCEPTION_LABEL_PEDESTRIAN",
    "PERCEPTION_LABEL_ANIMAL",
    "AVRESEARCH_LABEL_DONTCARE",
]
PERCEPTION_LABEL_TO_INDEX = {label: index for (index, label) in enumerate(PERCEPTION_LABELS)}

TL_FACE_LABELS = [
    "ACTIVE",
    "INACTIVE",
    "UNKNOWN",
]
TL_FACE_LABEL_TO_INDEX = {label: index for (index, label) in enumerate(TL_FACE_LABELS)}



import math
from pathlib import Path
from typing import NamedTuple

import numpy as np
import torch
from PIL import Image

from l5kit.rasterization import Rasterizer
from l5kit.simulation.dataset import SimulationDataset


class NonKinematicActionRescaleParams(NamedTuple):
    """Defines the parameters to rescale actions into un-normalized action space
    when the kinematic model is not used.

    :param x_mu: the translation of the x-coordinate
    :param x_scale: the scaling of the x-coordinate
    :param y_mu: the translation of the y-coordinate
    :param y_scale: the scaling of the y-coordinate
    :param yaw_mu: the translation of the yaw (radians)
    :param yaw_scale: the scaling of the yaw (radians)
    """
    x_mu: float
    x_scale: float
    y_mu: float
    y_scale: float
    yaw_mu: float
    yaw_scale: float


class KinematicActionRescaleParams(NamedTuple):
    """Defines the parameters to rescale actions into un-normalized action space
    when the kinematic model is used.

    :param steer_scale: the scaling of the steer (kinematic model)
    :param acc_scale: the scaling of the acceleration (kinematic model)
    """
    steer_scale: float
    acc_scale: float


def calculate_non_kinematic_rescale_params(sim_dataset: SimulationDataset) -> NonKinematicActionRescaleParams:
    """Calculate the action un-normalization parameters from the simulation dataset for non-kinematic model.

    :param sim_dataset: the input dataset to calculate the action rescale parameters
    :return: the unnormalized action
    """
    x_component_frames = []
    y_component_frames = []
    yaw_component_frames = []

    for index in range(1, len(sim_dataset) - 1):
        ego_input = sim_dataset.rasterise_frame_batch(index)
        x_component_frames.append([scene['target_positions'][0, 0] for scene in ego_input])
        y_component_frames.append([scene['target_positions'][0, 1] for scene in ego_input])
        yaw_component_frames.append([scene['target_yaws'][0, 0] for scene in ego_input])

    x_components = np.concatenate(x_component_frames)
    y_components = np.concatenate(y_component_frames)
    yaw_components = np.concatenate(yaw_component_frames)

    x_mu, x_std = np.mean(x_components), np.std(x_components)
    y_mu, y_std = np.mean(y_components), np.std(y_components)
    yaw_mu, yaw_std = np.mean(yaw_components), np.std(yaw_components)

    # Keeping scale = 10 * std so that extreme values are not clipped
    return NonKinematicActionRescaleParams(x_mu, 10 * x_std, y_mu, 10 * y_std, yaw_mu, 10 * yaw_std)


def calculate_kinematic_rescale_params(sim_dataset: SimulationDataset) -> KinematicActionRescaleParams:
    """Calculate the action un-normalization parameters from the simulation dataset for kinematic model.

    :param sim_dataset: the input dataset to calculate the action rescale parameters
    :return: the unnormalized action
    """
    v_component_frames = []
    yaw_component_frames = []

    for index in range(1, len(sim_dataset)):
        ego_input = sim_dataset.rasterise_frame_batch(index)
        v_component_frames.append([scene['curr_speed'].item() for scene in ego_input])
        yaw_component_frames.append([scene['target_yaws'][0, 0] for scene in ego_input])

    v_components = np.stack(v_component_frames)
    acc_components = v_components[1:] - v_components[:-1]
    acc_components = 0.5 * (acc_components[1:] + acc_components[:-1])
    acc_components = 0.5 * (acc_components[1:] + acc_components[:-1])
    acc_components = 0.5 * (acc_components[1:] + acc_components[:-1])

    # import pdb; pdb.set_trace()

    acc_components = acc_components.flatten()
    # acc_mu, acc_std = np.mean(acc_components), np.std(acc_components)

    yaw_components = np.concatenate(yaw_component_frames)
    # yaw_mu, yaw_std = np.mean(yaw_components), np.std(yaw_components)

    print(max(acc_components), min(acc_components))
    print(max(yaw_components), min(yaw_components), math.radians(20) * 0.1)
    # import pdb; pdb.set_trace()
    # assert max(acc_components) <= 0.7
    # assert -0.7 <= min(acc_components)

    # 15, 0.3
    return KinematicActionRescaleParams(math.radians(20) * 0.1, 0.6)


def save_input_raster(rasterizer: Rasterizer, image: torch.Tensor, num_images: int = 20,
                      output_folder: str = 'raster_inputs') -> None:
    """Save the input raster image.

    :param rasterizer: the rasterizer
    :param image: numpy array
    :param num_images: number of images to save
    :param output_folder: directory to save the image
    :return: the numpy dict with 'positions' and 'yaws'
    """

    image = image.permute(1, 2, 0).cpu().numpy()
    output_im = rasterizer.to_rgb(image)

    im = Image.fromarray(output_im)

    # mkdir
    Path(output_folder).mkdir(exist_ok=True)
    output_folder = Path(output_folder)

    # loop
    i = 0
    img_path = output_folder / f"input{i}.png"
    while img_path.exists():
        i += 1
        img_path = output_folder / f"input{i}.png"

    # save
    im.save(img_path)

    # exit code once num_images images saved
    if i == num_images:
        exit()



from typing import List, Tuple

import numpy as np

from l5kit.data import ChunkedDataset
from l5kit.data.filter import (filter_agents_by_frames, filter_agents_by_labels, filter_tl_faces_by_frames,
                               filter_tl_faces_by_status)
from l5kit.data.labels import PERCEPTION_LABELS
from l5kit.data.map_api import MapAPI, TLFacesColors
from l5kit.environment.envs.l5_env import EpisodeOutputGym
from l5kit.geometry import transform_points
from l5kit.rasterization.box_rasterizer import get_box_world_coords, get_ego_as_agent
from l5kit.rasterization.semantic_rasterizer import indices_in_bounds
from l5kit.sampling import get_relative_poses
from l5kit.simulation.unroll import SimulationOutput, UnrollInputOutput
from l5kit.visualization.visualizer.common import (AgentVisualization, CWVisualization, EgoVisualization,
                                                   FrameVisualization, LaneVisualization, TrajectoryVisualization)


# TODO: this should not be here (maybe a config?)
COLORS = {
    TLFacesColors.GREEN.name: "#33CC33",
    TLFacesColors.RED.name: "#FF3300",
    TLFacesColors.YELLOW.name: "#FFFF66",
    "PERCEPTION_LABEL_CAR": "#1F77B4",
    "PERCEPTION_LABEL_CYCLIST": "#CC33FF",
    "PERCEPTION_LABEL_PEDESTRIAN": "#66CCFF",
}


def _get_frame_trajectories(frames: np.ndarray, agents_frames: List[np.ndarray], track_ids: np.ndarray,
                            frame_index: int) -> List[TrajectoryVisualization]:
    """Get trajectories (ego and agents) starting at frame_index.
    Ego's trajectory will be named ego_trajectory while agents' agent_trajectory

    :param frames: all frames from the scene
    :param agents_frames: all agents from the scene as a list of array (one per frame)
    :param track_ids: allowed tracks ids we want to build trajectory for
    :param frame_index: index of the frame (trajectory will start from this frame)
    :return: a list of trajectory for visualisation
    """

    traj_visualisation: List[TrajectoryVisualization] = []
    # TODO: factor out future length
    agent_traj_length = 20
    for track_id in track_ids:
        # TODO this is not really relative (note eye and 0 yaw)
        pos, *_, avail = get_relative_poses(agent_traj_length, frames[frame_index: frame_index + agent_traj_length],
                                            track_id, agents_frames[frame_index: frame_index + agent_traj_length],
                                            np.eye(3), 0)
        traj_visualisation.append(TrajectoryVisualization(xs=pos[avail > 0, 0],
                                                          ys=pos[avail > 0, 1],
                                                          color="blue",
                                                          legend_label="agent_trajectory",
                                                          track_id=int(track_id)))

    # TODO: factor out future length
    ego_traj_length = 100
    pos, *_, avail = get_relative_poses(ego_traj_length, frames[frame_index: frame_index + ego_traj_length],
                                        None, agents_frames[frame_index: frame_index + ego_traj_length],
                                        np.eye(3), 0)
    traj_visualisation.append(TrajectoryVisualization(xs=pos[avail > 0, 0],
                                                      ys=pos[avail > 0, 1],
                                                      color="red",
                                                      legend_label="ego_trajectory",
                                                      track_id=-1))

    return traj_visualisation


def _get_frame_data(mapAPI: MapAPI, frame: np.ndarray, agents_frame: np.ndarray,
                    tls_frame: np.ndarray) -> FrameVisualization:
    """Get visualisation objects for the current frame.

    :param mapAPI: mapAPI object (used for lanes, crosswalks etc..)
    :param frame: the current frame (used for ego)
    :param agents_frame: agents in this frame
    :param tls_frame: the tls of this frame
    :return: A FrameVisualization object. NOTE: trajectory are not included here
    """
    ego_xy = frame["ego_translation"][:2]

    #################
    # plot lanes
    lane_indices = indices_in_bounds(ego_xy, mapAPI.bounds_info["lanes"]["bounds"], 50)
    active_tl_ids = set(filter_tl_faces_by_status(tls_frame, "ACTIVE")["face_id"].tolist())
    lanes_vis: List[LaneVisualization] = []

    for idx, lane_idx in enumerate(lane_indices):
        lane_idx = mapAPI.bounds_info["lanes"]["ids"][lane_idx]

        lane_tl_ids = set(mapAPI.get_lane_traffic_control_ids(lane_idx))
        lane_colour = "gray"
        for tl_id in lane_tl_ids.intersection(active_tl_ids):
            lane_colour = COLORS[mapAPI.get_color_for_face(tl_id)]

        lane_coords = mapAPI.get_lane_coords(lane_idx)
        left_lane = lane_coords["xyz_left"][:, :2]
        right_lane = lane_coords["xyz_right"][::-1, :2]

        lanes_vis.append(LaneVisualization(xs=np.hstack((left_lane[:, 0], right_lane[:, 0])),
                                           ys=np.hstack((left_lane[:, 1], right_lane[:, 1])),
                                           color=lane_colour))

    #################
    # plot crosswalks
    crosswalk_indices = indices_in_bounds(ego_xy, mapAPI.bounds_info["crosswalks"]["bounds"], 50)
    crosswalks_vis: List[CWVisualization] = []

    for idx in crosswalk_indices:
        crosswalk = mapAPI.get_crosswalk_coords(mapAPI.bounds_info["crosswalks"]["ids"][idx])
        crosswalks_vis.append(CWVisualization(xs=crosswalk["xyz"][:, 0],
                                              ys=crosswalk["xyz"][:, 1],
                                              color="yellow"))
    #################
    # plot ego and agents
    agents_frame = np.insert(agents_frame, 0, get_ego_as_agent(frame))
    box_world_coords = get_box_world_coords(agents_frame)

    # ego
    ego_vis = EgoVisualization(xs=box_world_coords[0, :, 0], ys=box_world_coords[0, :, 1],
                               color="red", center_x=agents_frame["centroid"][0, 0],
                               center_y=agents_frame["centroid"][0, 1])

    # agents
    agents_frame = agents_frame[1:]
    box_world_coords = box_world_coords[1:]

    agents_vis: List[AgentVisualization] = []
    for agent, box_coord in zip(agents_frame, box_world_coords):
        label_index = np.argmax(agent["label_probabilities"])
        agent_type = PERCEPTION_LABELS[label_index]
        agents_vis.append(AgentVisualization(xs=box_coord[..., 0],
                                             ys=box_coord[..., 1],
                                             color="#1F77B4" if agent_type not in COLORS else COLORS[agent_type],
                                             track_id=agent["track_id"],
                                             agent_type=PERCEPTION_LABELS[label_index],
                                             prob=agent["label_probabilities"][label_index]))

    return FrameVisualization(ego=ego_vis, agents=agents_vis, lanes=lanes_vis,
                              crosswalks=crosswalks_vis, trajectories=[])


def zarr_to_visualizer_scene(scene_dataset: ChunkedDataset, mapAPI: MapAPI,
                             with_trajectories: bool = True) -> List[FrameVisualization]:
    """Convert a zarr scene into a list of FrameVisualization which can be used by the visualiser

    :param scene_dataset: a scene dataset. This must contain a single scene
    :param mapAPI: mapAPI object
    :param with_trajectories: if to enable trajectories or not
    :return: a list of FrameVisualization objects
    """
    if len(scene_dataset.scenes) != 1:
        raise ValueError(f"we can convert only a single scene, found {len(scene_dataset.scenes)}")

    frames = scene_dataset.frames
    agents_frames = filter_agents_by_frames(frames, scene_dataset.agents)
    tls_frames = filter_tl_faces_by_frames(frames, scene_dataset.tl_faces)

    frames_vis: List[FrameVisualization] = []
    for frame_idx in range(len(frames)):
        frame = frames[frame_idx]
        tls_frame = tls_frames[frame_idx]

        # TODO: hardcoded threshold, it would be great to have a slider filtering on this
        agents_frame = agents_frames[frame_idx]
        agents_frame = filter_agents_by_labels(agents_frame, 0.1)

        frame_vis = _get_frame_data(mapAPI, frame, agents_frame, tls_frame)

        if with_trajectories:
            traj_vis = _get_frame_trajectories(frames, agents_frames, agents_frame["track_id"], frame_idx)
            frame_vis = FrameVisualization(ego=frame_vis.ego, agents=frame_vis.agents,
                                           lanes=frame_vis.lanes, crosswalks=frame_vis.crosswalks,
                                           trajectories=traj_vis)
        frames_vis.append(frame_vis)

    return frames_vis


def _get_in_out_as_trajectories(in_out: UnrollInputOutput) -> Tuple[np.ndarray, np.ndarray]:
    """Convert the input (log-replayed) and output (simulated) trajectories into world space.
    Apply availability on the log-replayed one

    :param in_out: an UnrollInputOutput object
    :return: the replayed and simulated trajectory as numpy arrays
    """
    replay_traj = transform_points(in_out.inputs["target_positions"],
                                   in_out.inputs["world_from_agent"])
    replay_traj = replay_traj[in_out.inputs["target_availabilities"] > 0]
    sim_traj = transform_points(in_out.outputs["positions"],
                                in_out.inputs["world_from_agent"])

    return replay_traj, sim_traj


def simulation_out_to_visualizer_scene(sim_out: SimulationOutput, mapAPI: MapAPI) -> List[FrameVisualization]:
    """Convert a simulation output into a scene we can visualize.
    The scene will include replayed and simulated trajectories for ego and agents when these are
    simulated.

    :param sim_out: the simulation output
    :param mapAPI: a MapAPI object
    :return: a list of FrameVisualization for the scene
    """
    frames = sim_out.simulated_ego
    agents_frames = filter_agents_by_frames(frames, sim_out.simulated_agents)
    tls_frames = filter_tl_faces_by_frames(frames, sim_out.simulated_dataset.dataset.tl_faces)
    agents_th = sim_out.simulated_dataset.cfg["raster_params"]["filter_agents_threshold"]
    ego_ins_outs = sim_out.ego_ins_outs
    agents_ins_outs = sim_out.agents_ins_outs

    has_ego_info = len(ego_ins_outs) > 0
    has_agents_info = len(agents_ins_outs) > 0

    frames_vis: List[FrameVisualization] = []
    for frame_idx in range(len(frames)):
        frame = frames[frame_idx]
        tls_frame = tls_frames[frame_idx]

        agents_frame = agents_frames[frame_idx]
        agents_frame = filter_agents_by_labels(agents_frame, agents_th)
        frame_vis = _get_frame_data(mapAPI, frame, agents_frame, tls_frame)
        trajectories = []

        if has_ego_info:
            ego_in_out = ego_ins_outs[frame_idx]
            replay_traj, sim_traj = _get_in_out_as_trajectories(ego_in_out)
            trajectories.append(TrajectoryVisualization(xs=replay_traj[:, 0], ys=replay_traj[:, 1],
                                                        color="blue", legend_label="ego_replay", track_id=-1))
            trajectories.append(TrajectoryVisualization(xs=sim_traj[:, 0], ys=sim_traj[:, 1],
                                                        color="red", legend_label="ego_simulated", track_id=-1))

        if has_agents_info:
            agents_in_out = agents_ins_outs[frame_idx]
            for agent_in_out in agents_in_out:
                track_id = agent_in_out.inputs["track_id"]
                replay_traj, sim_traj = _get_in_out_as_trajectories(agent_in_out)
                trajectories.append(TrajectoryVisualization(xs=replay_traj[:, 0], ys=replay_traj[:, 1],
                                                            color="orange", legend_label="agent_replay",
                                                            track_id=track_id))
                trajectories.append(TrajectoryVisualization(xs=sim_traj[:, 0], ys=sim_traj[:, 1],
                                                            color="purple", legend_label="agent_simulated",
                                                            track_id=track_id))

        frame_vis = FrameVisualization(ego=frame_vis.ego, agents=frame_vis.agents,
                                       lanes=frame_vis.lanes, crosswalks=frame_vis.crosswalks,
                                       trajectories=trajectories)

        frames_vis.append(frame_vis)

    return frames_vis


def episode_out_to_visualizer_scene_gym_cle(sim_out: EpisodeOutputGym,
                                            mapAPI: MapAPI) -> List[FrameVisualization]:
    """Convert a episode output of closed loop gym into a scene we can visualize.
    The scene will include replayed and simulated trajectories for ego and agents when these are
    simulated.

    :param sim_out: the simulation output of L5 gym close loop environment
    :param mapAPI: a MapAPI object
    :return: a list of FrameVisualization for the scene
    """
    frames = sim_out.simulated_ego
    agents_frames = filter_agents_by_frames(frames, sim_out.simulated_agents)
    tls_frames = filter_tl_faces_by_frames(frames, sim_out.tls_frames)
    agents_th = sim_out.agents_th
    ego_ins_outs = sim_out.ego_ins_outs
    agents_ins_outs = sim_out.agents_ins_outs

    has_ego_info = len(ego_ins_outs) > 0
    has_agents_info = len(agents_ins_outs) > 0

    frames_vis: List[FrameVisualization] = []
    for frame_idx in range(len(frames) - 2):
        frame = frames[frame_idx]
        tls_frame = tls_frames[frame_idx]

        agents_frame = agents_frames[frame_idx]
        agents_frame = filter_agents_by_labels(agents_frame, agents_th)
        frame_vis = _get_frame_data(mapAPI, frame, agents_frame, tls_frame)
        trajectories = []

        if has_ego_info:
            ego_in_out = ego_ins_outs[frame_idx]
            replay_traj, sim_traj = _get_in_out_as_trajectories(ego_in_out)
            scale = 10
            ego_centroid = np.array([[frame_vis.ego.center_x, frame_vis.ego.center_y]])
            ego_next_step_replay = ego_centroid + scale * (replay_traj[0:1] - ego_centroid)
            ego_next_step_sim = ego_centroid + scale * (sim_traj[0:1] - ego_centroid)
            single_step_replay = np.concatenate([ego_centroid, ego_next_step_replay])
            single_step_sim = np.concatenate([ego_centroid, ego_next_step_sim])

            trajectories.append(TrajectoryVisualization(xs=single_step_sim[:, 0], ys=single_step_sim[:, 1],
                                                        color="green", legend_label="ego_simulated (10x scale)",
                                                        track_id=-1))
            trajectories.append(TrajectoryVisualization(xs=single_step_replay[:, 0], ys=single_step_replay[:, 1],
                                                        color="yellow", legend_label="ego_replay (10x scale)",
                                                        track_id=-1))

        if has_agents_info:
            agents_in_out = agents_ins_outs[frame_idx]
            for agent_in_out in agents_in_out:
                track_id = agent_in_out.inputs["track_id"]
                replay_traj, sim_traj = _get_in_out_as_trajectories(agent_in_out)
                trajectories.append(TrajectoryVisualization(xs=replay_traj[:, 0], ys=replay_traj[:, 1],
                                                            color="orange", legend_label="agent_replay",
                                                            track_id=track_id))
                trajectories.append(TrajectoryVisualization(xs=sim_traj[:, 0], ys=sim_traj[:, 1],
                                                            color="purple", legend_label="agent_simulated",
                                                            track_id=track_id))

        frame_vis = FrameVisualization(ego=frame_vis.ego, agents=frame_vis.agents,
                                       lanes=frame_vis.lanes, crosswalks=frame_vis.crosswalks,
                                       trajectories=trajectories)

        frames_vis.append(frame_vis)

    return frames_vis



def get_history_slice(
        frame_index: int, history_num_states: int, history_step_size: int, include_current_state: bool = False
) -> slice:
    """Given a frame index and history settings returns a slice that returns the given data in the right order.
    Note that this history returned starts with the most "recent" frame first (i.e. reverse in time as it's history).

    Example:
    ``frame_index=20``, ``history_num_frames=2``, ``history_step_size=2``, ``include_current_state=True``
    would return a slice for frame index 20, 18, 16.

    Arguments:
        state_index (int): The "anchor" frame index you want to sample from
        history_num_states (int): Number of history frames (not including the current frame).
        history_step_size (int): How many frames to step for each history step.

    Keyword Arguments:
        include_current_state (bool): Whether the slice should include ``frame_index`` (default: {False})

    Raises:
        IndexError: Returned when ``history_step_size`` is an invalid value (e.g. 0).

    Returns:
        slice -- Slice that when applied to an array returns the history frames in the right order.
    """
    if history_step_size <= 0:
        raise IndexError("History step size can not be 0 or negative")
    history_earliest_index = frame_index - (history_num_states) * history_step_size - 1
    history_latest_index = frame_index - (0 if include_current_state else history_step_size)

    if history_latest_index < 0:
        return slice(0, 0, -history_step_size)

    # Necessary for including the first element with negative step size (step size has to be negated as it's history)
    # Example: start index = 2, step size = 2, element 0 should be included.
    # +1 is required because it is non-inclusive (-1 is performed above).
    if history_earliest_index < 0 and (history_earliest_index + 1) % history_step_size == 0:
        history_earliest_index = None  # type: ignore
    else:
        history_earliest_index = max(0, history_earliest_index)

    return slice(history_latest_index, history_earliest_index, -history_step_size)


def get_future_slice(frame_index: int, future_num_states: int, future_step_size: int) -> slice:
    """Given a frame index and future settings returns a slice that returns the given data in the right order.
    Note that this history returned starts with the most "recent" frame first
    (e.g. ``current_frame``+``future_step_size``).

    Example:
    ``frame_index=20``, ``future_num_states=2``, ``future_step_size=2``
    would return a slice for frame index 22, 24.

    Arguments:
        state_index (int): The "anchor" frame index you want to sample from
        future_num_states (int): Number of future frames.
        future_step_size (int): How many frames to step for each future step.

    Raises:
        IndexError: Returned when ``future_step_size`` is an invalid value (e.g. 0).

    Returns:
        slice -- Slice that when applied to an array returns the future frames in the right order.
    """
    if future_step_size <= 0:
        raise IndexError("Future step size can not be 0")

    future_latest_index = frame_index + (future_num_states) * future_step_size + 1
    future_earliest_index = frame_index + future_step_size
    return slice(future_earliest_index, future_latest_index, future_step_size)



import warnings
from typing import Tuple

import numpy as np

from ..geometry import rotation33_as_yaw, yaw_as_rotation33
from ..random import RandomGenerator
from .ackerman_steering_model import fit_ackerman_model_exact
from .perturbation import Perturbation


def _get_trajectory(past_frames: np.ndarray, future_frames: np.ndarray) -> np.ndarray:
    """A function that takes past & future frames, extracts ego translation & rotation from the frames, sorts them
    based on time (from past to future) then return sorted ego translation & rotation in a trajectory array.
    :param past_frames: a list of past frames
    :type past_frames: np.ndarray
    :param future_frames: a list of future frames
    :type future_frames: np.ndarray
    :return: a ego trajectory (position + heading) sorted from past to future
    :rtype: np.ndarray
    """
    num_frames = len(past_frames) + len(future_frames)
    trajectory = np.zeros((num_frames, 3), dtype=np.float32)

    # Note that history frames go backward in time from the anchor frame.
    trajectory[:, :2] = np.concatenate(
        (past_frames["ego_translation"][::-1, :2], future_frames["ego_translation"][:, :2])
    )
    rotations = np.concatenate((past_frames["ego_rotation"][::-1], future_frames["ego_rotation"]))
    trajectory[:, 2] = [rotation33_as_yaw(rot) for rot in rotations]

    return trajectory


class AckermanPerturbation(Perturbation):
    def __init__(
        self,
        random_offset_generator: RandomGenerator,
        perturb_prob: float,
        min_displacement: float = 0,
    ):
        """
        Apply Ackerman to get a feasible trajectory with probability perturb_prob.

        :param random_offset_generator: a callable that yields 3 values (lat, long, yaw offsets)
        :param perturb_prob: probability between 0 and 1 of applying the perturbation
        :param min_displacement: minimum displacement required to apply lateral & yaw perturbation
        """
        self.perturb_prob = perturb_prob
        self.min_displacement = min_displacement
        self.random_offset_generator = random_offset_generator
        if perturb_prob == 0:
            warnings.warn(
                "Consider replacing this object with None if no perturbation is intended", RuntimeWarning, stacklevel=2
            )

    def perturb(self, history_frames: np.ndarray, future_frames: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        if np.random.rand() >= self.perturb_prob:
            return history_frames.copy(), future_frames.copy()

        (
            lateral_offset_m,
            longitudinal_offset_m,
            yaw_offset_rad,
        ) = self.random_offset_generator()

        num_history_frames = len(history_frames)
        num_future_frames = len(future_frames)

        assert num_history_frames > 0, f"Number of history frames: {num_history_frames}"
        assert num_future_frames > 0, f"Number of future frames: {num_future_frames}"

        trajectory = _get_trajectory(history_frames, future_frames)

        curr_frame_idx = num_history_frames - 1
        displacements = np.linalg.norm(np.diff(trajectory[curr_frame_idx:, :2], axis=0), axis=1)

        # TODO: ackerman lateral & yaw perturbation does not work when EGO slow moving
        if np.sum(displacements) < self.min_displacement:
            lateral_offset_m = 0
            yaw_offset_rad = 0

        position_offset_m = np.array([longitudinal_offset_m, lateral_offset_m, 0])
        position_offset_m = np.matmul(yaw_as_rotation33(trajectory[curr_frame_idx, 2]), position_offset_m)

        #  perform ackerman steering model fitting
        gx = trajectory[curr_frame_idx + 1:, 0]
        gy = trajectory[curr_frame_idx + 1:, 1]
        gr = trajectory[curr_frame_idx + 1:, 2]
        gv = displacements

        x0 = trajectory[curr_frame_idx, 0] + position_offset_m[0]
        y0 = trajectory[curr_frame_idx, 1] + position_offset_m[1]
        r0 = trajectory[curr_frame_idx, 2] + yaw_offset_rad
        v0 = gv[0]

        wgx = np.ones(num_future_frames)
        wgy = np.ones(num_future_frames)
        wgr = np.ones(num_future_frames)
        wgv = np.zeros(num_future_frames)

        new_xs, new_ys, new_yaws, new_vs, new_acc, new_steer = fit_ackerman_model_exact(
            x0, y0, r0, v0, gx, gy, gr, gv, wgx, wgy, wgr, wgv,
        )

        history_frames["ego_translation"][0, 0] = x0
        history_frames["ego_translation"][0, 1] = y0
        history_frames["ego_rotation"][0] = yaw_as_rotation33(r0)

        future_frames["ego_translation"][:, 0] = new_xs
        future_frames["ego_translation"][:, 1] = new_ys
        future_frames["ego_rotation"] = np.array([yaw_as_rotation33(yaw) for yaw in new_yaws])

        return history_frames, future_frames
